// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System.Collections.Specialized;
using System.Text;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents an authentication challenge in the EonaCat network library.
    /// </summary>
    internal class AuthenticationChallenge : AuthenticationBase
    {
        private const string BASIC = "basic";
        private const string DIGEST = "digest";
        private const int DIGEST_SIZE = 128;

        /// <summary>
        /// Initializes a new instance of the <see cref="AuthenticationChallenge"/> class.
        /// </summary>
        /// <param name="scheme">The authentication scheme.</param>
        /// <param name="parameters">The collection of authentication parameters.</param>
        private AuthenticationChallenge(AuthenticationSchemes scheme, NameValueCollection parameters)
          : base(scheme, parameters)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="AuthenticationChallenge"/> class for Basic or Digest authentication.
        /// </summary>
        /// <param name="scheme">The authentication scheme.</param>
        /// <param name="realm">The authentication realm.</param>
        internal AuthenticationChallenge(AuthenticationSchemes scheme, string realm)
          : base(scheme, new NameValueCollection())
        {
            Parameters["realm"] = realm;
            if (scheme == AuthenticationSchemes.Digest)
            {
                Parameters["nonce"] = CreateNonceValue();
                Parameters["algorithm"] = "MD5";
                Parameters["qop"] = "auth";
            }
        }

        /// <summary>
        /// Gets the domain for Digest authentication.
        /// </summary>
        public string Domain => Parameters["domain"];

        /// <summary>
        /// Gets the stale parameter for Digest authentication.
        /// </summary>
        public string Stale => Parameters["stale"];

        /// <summary>
        /// Creates a Basic authentication challenge with the specified realm.
        /// </summary>
        /// <param name="realm">The authentication realm.</param>
        /// <returns>An instance of <see cref="AuthenticationChallenge"/> for Basic authentication.</returns>

        internal static AuthenticationChallenge CreateBasicChallenge(string realm)
        {
            return new AuthenticationChallenge(AuthenticationSchemes.Basic, realm);
        }

        /// <summary>
        /// Creates a Digest authentication challenge with the specified realm.
        /// </summary>
        /// <param name="realm">The authentication realm.</param>
        /// <returns>An instance of <see cref="AuthenticationChallenge"/> for Digest authentication.</returns>
        internal static AuthenticationChallenge CreateDigestChallenge(string realm)
        {
            return new AuthenticationChallenge(AuthenticationSchemes.Digest, realm);
        }

        /// <summary>
        /// Parses an authentication challenge from the specified string value.
        /// </summary>
        /// <param name="value">The string containing the authentication challenge.</param>
        /// <returns>An instance of <see cref="AuthenticationChallenge"/> if successful; otherwise, null.</returns>
        internal static AuthenticationChallenge Parse(string value)
        {
            var challenge = value.Split(new[] { ' ' }, 2);
            if (challenge.Length != 2)
            {
                return null;
            }

            var scheme = challenge[0].ToLower();
            return scheme == BASIC
                   ? new AuthenticationChallenge(
                       AuthenticationSchemes.Basic, ParseParameters(challenge[1]))
                   : scheme == DIGEST
                     ? new AuthenticationChallenge(
                         AuthenticationSchemes.Digest, ParseParameters(challenge[1]))
                     : null;
        }

        /// <summary>
        /// Gets the Basic authentication string representation of the authentication challenge.
        /// </summary>
        /// <returns>A string representing the Basic authentication challenge.</returns>
        internal override string ToBasicString()
        {
            return string.Format($"Basic realm=\"{{0}}\"", Parameters["realm"]);
        }

        /// <summary>
        /// Gets the Digest authentication string representation of the authentication challenge.
        /// </summary>
        /// <returns>A string representing the Digest authentication challenge.</returns>
        internal override string ToDigestString()
        {
            var output = new StringBuilder(DIGEST_SIZE);

            var domain = Parameters["domain"];
            if (domain != null)
            {
                output.AppendFormat($"Digest realm=\"{Parameters["realm"]}\", domain=\"{domain}\", nonce=\"{Parameters["nonce"]}\"");
            }
            else
            {
                output.AppendFormat($"Digest realm=\"{Parameters["realm"]}\", nonce=\"{Parameters["nonce"]}\"");
            }

            var opaque = Parameters["opaque"];
            if (opaque != null)
            {
                output.AppendFormat($", opaque=\"{opaque}\"");
            }

            var stale = Parameters["stale"];
            if (stale != null)
            {
                output.AppendFormat($", stale={stale}");
            }

            var algorithm = Parameters["algorithm"];
            if (algorithm != null)
            {
                output.AppendFormat($", algorithm={algorithm}");
            }

            var qop = Parameters["qop"];
            if (qop != null)
            {
                output.AppendFormat($", qop=\"{qop}\"");
            }

            return output.ToString();
        }
    }
}