// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections.Specialized;
using System.Security.Cryptography;
using System.Security.Principal;
using System.Text;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents an authentication response in the EonaCat network library.
    /// </summary>
    internal class AuthenticationResponse : AuthenticationBase
    {
        private const string BASIC = "basic";
        private const string DIGEST = "digest";
        private uint _nonceCount;

        /// <summary>
        /// Initializes a new instance of the <see cref="AuthenticationResponse"/> class.
        /// </summary>
        /// <param name="scheme">The authentication scheme.</param>
        /// <param name="parameters">The collection of authentication parameters.</param>
        private AuthenticationResponse(AuthenticationSchemes scheme, NameValueCollection parameters)
          : base(scheme, parameters)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="AuthenticationResponse"/> class for Basic authentication.
        /// </summary>
        /// <param name="credentials">The network credentials.</param>
        internal AuthenticationResponse(NetworkCredential credentials)
          : this(AuthenticationSchemes.Basic, new NameValueCollection(), credentials, 0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="AuthenticationResponse"/> class for Digest authentication.
        /// </summary>
        /// <param name="challenge">The authentication challenge.</param>
        /// <param name="credentials">The network credentials.</param>
        /// <param name="nonceCount">The nonce count.</param>
        internal AuthenticationResponse(
          AuthenticationChallenge challenge, NetworkCredential credentials, uint nonceCount)
          : this(challenge.Scheme, challenge.Parameters, credentials, nonceCount)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="AuthenticationResponse"/> class.
        /// </summary>
        /// <param name="scheme">The authentication scheme.</param>
        /// <param name="parameters">The collection of authentication parameters.</param>
        /// <param name="credentials">The network credentials.</param>
        /// <param name="nonceCount">The nonce count.</param>
        internal AuthenticationResponse(
          AuthenticationSchemes scheme,
          NameValueCollection parameters,
          NetworkCredential credentials,
          uint nonceCount)
          : base(scheme, parameters)
        {
            Parameters["username"] = credentials.Username;
            Parameters["password"] = credentials.Password;
            Parameters["uri"] = credentials.Domain;
            _nonceCount = nonceCount;
            if (scheme == AuthenticationSchemes.Digest)
            {
                InitAsDigest();
            }
        }

        /// <summary>
        /// Gets the nonce count.
        /// </summary>
        internal uint NonceCount => _nonceCount < uint.MaxValue
                       ? _nonceCount
                       : 0;

        /// <summary>
        /// Gets the cnonce value for Digest authentication.
        /// </summary>
        public string Cnonce => Parameters["cnonce"];

        /// <summary>
        /// Gets the nonce count for Digest authentication.
        /// </summary>
        public string Nc => Parameters["nc"];

        /// <summary>
        /// Gets the password for authentication.
        /// </summary>
        public string Password => Parameters["password"];

        /// <summary>
        /// Gets the response value for authentication.
        /// </summary>
        public string Response => Parameters["response"];

        /// <summary>
        /// Gets the URI for authentication.
        /// </summary>
        public string Uri => Parameters["uri"];

        /// <summary>
        /// Gets the username for authentication.
        /// </summary>
        public string UserName => Parameters["username"];

        /// <summary>
        /// Creates the A1 value for Digest authentication.
        /// </summary>
        /// <param name="username">The username.</param>
        /// <param name="password">The password.</param>
        /// <param name="realm">The realm.</param>
        /// <returns>The A1 value.</returns>
        private static string CreateA1(string username, string password, string realm)
        {
            return $"{username}:{realm}:{password}";
        }

        /// <summary>
        /// Creates the A1 value for Digest authentication with cnonce and nonce.
        /// </summary>
        /// <param name="username">The username.</param>
        /// <param name="password">The password.</param>
        /// <param name="realm">The realm.</param>
        /// <param name="nonce">The nonce.</param>
        /// <param name="cnonce">The cnonce.</param>
        /// <returns>The A1 value.</returns>
        private static string CreateA1(
          string username, string password, string realm, string nonce, string cnonce)
        {
            return $"{Hash(CreateA1(username, password, realm))}:{nonce}:{cnonce}";
        }

        /// <summary>
        /// Creates the A2 value for Digest authentication.
        /// </summary>
        /// <param name="method">The HTTP method.</param>
        /// <param name="uri">The URI.</param>
        /// <returns>The A2 value.</returns>
        private static string CreateA2(string method, string uri)
        {
            return $"{method}:{uri}";
        }

        /// <summary>
        /// Creates the A2 value for Digest authentication with an entity.
        /// </summary>
        /// <param name="method">The HTTP method.</param>
        /// <param name="uri">The URI.</param>
        /// <param name="entity">The entity.</param>
        /// <returns>The A2 value.</returns>
        private static string CreateA2(string method, string uri, string entity)
        {
            return $"{method}:{uri}:{Hash(entity)}";
        }

        /// <summary>
        /// Computes the MD5 hash of the given value.
        /// </summary>
        /// <param name="value">The value to hash.</param>
        /// <returns>The MD5 hash.</returns>
        private static string Hash(string value)
        {
            var source = Encoding.UTF8.GetBytes(value);
            var md5 = MD5.Create();
            var hashed = md5.ComputeHash(source);

            var result = new StringBuilder(64);
            foreach (var currentByte in hashed)
            {
                result.Append(currentByte.ToString("x2"));
            }

            return result.ToString();
        }

        /// <summary>
        /// Initializes the authentication as Digest.
        /// </summary>
        private void InitAsDigest()
        {
            var qops = Parameters["qop"];
            if (qops != null)
            {
                if (qops.Split(',').Contains(qop => qop.Trim().ToLower() == "auth"))
                {
                    Parameters["qop"] = "auth";
                    Parameters["cnonce"] = CreateNonceValue();
                    Parameters["nc"] = string.Format("{0:x8}", ++_nonceCount);
                }
                else
                {
                    Parameters["qop"] = null;
                }
            }

            Parameters["method"] = "GET";
            Parameters["response"] = CreateRequestDigest(Parameters);
        }

        /// <summary>
        /// Creates the request digest for Digest authentication.
        /// </summary>
        /// <param name="parameters">The authentication parameters.</param>
        /// <returns>The request digest.</returns>
        internal static string CreateRequestDigest(NameValueCollection parameters)
        {
            var user = parameters["username"];
            var pass = parameters["password"];
            var realm = parameters["realm"];
            var nonce = parameters["nonce"];
            var uri = parameters["uri"];
            var algo = parameters["algorithm"];
            var qop = parameters["qop"];
            var cnonce = parameters["cnonce"];
            var nc = parameters["nc"];
            var method = parameters["method"];

            var a1 = algo != null && algo.ToLower() == "md5-sess"
                     ? CreateA1(user, pass, realm, nonce, cnonce)
                     : CreateA1(user, pass, realm);

            var a2 = qop != null && qop.ToLower() == "auth-int"
                     ? CreateA2(method, uri, parameters["entity"])
                     : CreateA2(method, uri);

            var secret = Hash(a1);
            var data = qop != null
                       ? string.Format("{0}:{1}:{2}:{3}:{4}", nonce, nc, cnonce, qop, Hash(a2))
                       : string.Format("{0}:{1}", nonce, Hash(a2));

            return Hash(string.Format("{0}:{1}", secret, data));
        }

        /// <summary>
        /// Parses an authentication response from the specified string value.
        /// </summary>
        /// <param name="value">The string containing the authentication response.</param>
        /// <returns>An instance of <see cref="AuthenticationResponse"/> if successful; otherwise, null.</returns>
        internal static AuthenticationResponse Parse(string value)
        {
            try
            {
                var cred = value.Split(new[] { ' ' }, 2);
                if (cred.Length != 2)
                {
                    return null;
                }

                var scheme = cred[0].ToLower();
                return scheme == BASIC
                       ? new AuthenticationResponse(
                           AuthenticationSchemes.Basic, ParseBasicCredentials(cred[1]))
                       : scheme == DIGEST
                         ? new AuthenticationResponse(
                             AuthenticationSchemes.Digest, ParseParameters(cred[1]))
                         : null;
            }
            catch
            {
            }

            return null;
        }

        /// <summary>
        /// Parses the basic credentials from the specified string value.
        /// </summary>
        /// <param name="value">The string containing basic credentials.</param>
        /// <returns>A collection of basic credentials.</returns>
        internal static NameValueCollection ParseBasicCredentials(string value)
        {
            // Decode the basic-credentials (a Base64 encoded string).
            var userPass = Encoding.Default.GetString(Convert.FromBase64String(value));

            // The format is [<domain>\]<username>:<password>.
            var i = userPass.IndexOf(':');
            var user = userPass.Substring(0, i);
            var pass = i < userPass.Length - 1 ? userPass.Substring(i + 1) : string.Empty;

            // Check if 'domain' exists.
            i = user.IndexOf('\\');
            if (i > -1)
            {
                user = user.Substring(i + 1);
            }

            var res = new NameValueCollection
            {
                ["username"] = user,
                ["password"] = pass
            };

            return res;
        }

        /// <summary>
        /// Gets the Basic authentication string representation of the authentication response.
        /// </summary>
        /// <returns>A string representing the Basic authentication response.</returns>
        internal override string ToBasicString()
        {
            var userPass = string.Format("{0}:{1}", Parameters["username"], Parameters["password"]);
            var cred = Convert.ToBase64String(Encoding.UTF8.GetBytes(userPass));

            return "Basic " + cred;
        }

        /// <summary>
        /// Gets the Digest authentication string representation of the authentication response.
        /// </summary>
        /// <returns>A string representing the Digest authentication response.</returns>
        internal override string ToDigestString()
        {
            var output = new StringBuilder(256);
            output.AppendFormat($"Digest username=\"{Parameters["username"]}\", realm=\"{Parameters["realm"]}\", nonce=\"{Parameters["nonce"]}\", uri=\"{Parameters["uri"]}\", response=\"{Parameters["response"]}\"");

            var opaque = Parameters["opaque"];
            if (opaque != null)
            {
                output.AppendFormat($", opaque=\"{opaque}\"");
            }

            var algorithm = Parameters["algorithm"];
            if (algorithm != null)
            {
                output.AppendFormat($", algorithm={algorithm}");
            }

            var qop = Parameters["qop"];
            if (qop != null)
            {
                output.AppendFormat($", qop={qop}, cnonce=\"{Parameters["cnonce"]}\", nc={Parameters["nc"]}");
            }

            return output.ToString();
        }

        /// <summary>
        /// Converts the authentication response to an identity.
        /// </summary>
        /// <returns>An instance of <see cref="IIdentity"/>.</returns>
        public IIdentity ToIdentity()
        {
            var scheme = Scheme;
            return scheme == AuthenticationSchemes.Basic
                   ? new HttpBasicIdentity(Parameters["username"], Parameters["password"])
                   : scheme == AuthenticationSchemes.Digest
                     ? new HttpDigestIdentity(Parameters)
                     : null;
        }
    }
}