// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents network credentials used for authentication in the EonaCat network library.
    /// </summary>
    public class NetworkCredential
    {
        private string _domain;
        private static readonly string[] _noRoles;
        private string _password;
        private string[] _roles;

        static NetworkCredential()
        {
            _noRoles = new string[0];
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="NetworkCredential"/> class with the specified username and password.
        /// </summary>
        /// <param name="username">The username.</param>
        /// <param name="password">The password.</param>
        public NetworkCredential(string username, string password)
          : this(username, password, null, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="NetworkCredential"/> class with the specified username, password, domain, and roles.
        /// </summary>
        /// <param name="username">The username.</param>
        /// <param name="password">The password.</param>
        /// <param name="domain">The domain.</param>
        /// <param name="roles">The roles.</param>
        public NetworkCredential(
          string username, string password, string domain, params string[] roles
        )
        {
            if (username == null)
            {
                throw new ArgumentNullException(nameof(username));
            }

            if (username.Length == 0)
            {
                throw new ArgumentException("An empty string.", nameof(username));
            }

            Username = username;
            _password = password;
            _domain = domain;
            _roles = roles;
        }

        /// <summary>
        /// Gets or sets the domain associated with the network credentials.
        /// </summary>
        public string Domain
        {
            get
            {
                return _domain ?? string.Empty;
            }

            internal set
            {
                _domain = value;
            }
        }

        /// <summary>
        /// Gets or sets the password associated with the network credentials.
        /// </summary>
        public string Password
        {
            get
            {
                return _password ?? string.Empty;
            }

            internal set
            {
                _password = value;
            }
        }

        /// <summary>
        /// Gets or sets the roles associated with the network credentials.
        /// </summary>
        public string[] Roles
        {
            get
            {
                return _roles ?? _noRoles;
            }

            internal set
            {
                _roles = value;
            }
        }

        /// <summary>
        /// Gets the username associated with the network credentials.
        /// </summary>
        public string Username { get; internal set; }
    }
}