// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Net;
using System.Text;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents a stream for handling chunked data in the EonaCat network library.
    /// </summary>
    internal class ChunkStream
    {
        private int _chunkRead;
        private int _chunkSize;
        private readonly List<WebChunk> _chunks;
        private bool _foundSPCode;
        private readonly StringBuilder _saved;
        private bool _gotChunck;
        private InputChunkState _state;
        private int _trailerState;

        /// <summary>
        /// Initializes a new instance of the <see cref="ChunkStream"/> class with the specified headers.
        /// </summary>
        /// <param name="headers">The web headers associated with the chunk stream.</param>
        public ChunkStream(WebHeaderCollection headers)
        {
            Headers = headers;
            _chunkSize = -1;
            _chunks = new List<WebChunk>();
            _saved = new StringBuilder();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ChunkStream"/> class with the specified buffer, offset, count, and headers.
        /// </summary>
        /// <param name="buffer">The byte array containing chunked data.</param>
        /// <param name="offset">The offset in the buffer at which to begin reading.</param>
        /// <param name="count">The number of bytes to read from the buffer.</param>
        /// <param name="headers">The web headers associated with the chunk stream.</param>
        public ChunkStream(byte[] buffer, int offset, int count, WebHeaderCollection headers)
          : this(headers)
        {
            Write(buffer, offset, count);
        }

        /// <summary>
        /// Gets the web headers associated with the chunk stream.
        /// </summary>
        internal WebHeaderCollection Headers { get; }

        /// <summary>
        /// Gets the number of bytes left in the current chunk.
        /// </summary>
        public int ChunkLeft => _chunkSize - _chunkRead;

        /// <summary>
        /// Gets a value indicating whether more data is expected.
        /// </summary>
        public bool WantMore => _state != InputChunkState.End;

        private int read(byte[] buffer, int offset, int count)
        {
            var nread = 0;

            var cnt = _chunks.Count;
            for (var i = 0; i < cnt; i++)
            {
                var chunk = _chunks[i];
                if (chunk == null)
                {
                    continue;
                }

                if (chunk.ReadLeft == 0)
                {
                    _chunks[i] = null;
                    continue;
                }

                nread += chunk.Read(buffer, offset + nread, count - nread);
                if (nread == count)
                {
                    break;
                }
            }

            return nread;
        }

        private static string RemoveChunkExtension(string value)
        {
            var index = value.IndexOf(';');
            return index > -1 ? value.Substring(0, index) : value;
        }

        private InputChunkState SeekCrLf(byte[] buffer, ref int offset, int length)
        {
            if (!_gotChunck)
            {
                if (buffer[offset++] != 13)
                {
                    ThrowProtocolViolation("CR is expected.");
                }

                _gotChunck = true;
                if (offset == length)
                {
                    return InputChunkState.DataEnded;
                }
            }

            if (buffer[offset++] != 10)
            {
                ThrowProtocolViolation("LF is expected.");
            }

            return InputChunkState.None;
        }

        private InputChunkState SetChunkSize(byte[] buffer, ref int offset, int length)
        {
            byte currentByte = 0;
            while (offset < length)
            {
                currentByte = buffer[offset++];
                if (_gotChunck)
                {
                    if (currentByte != 10)
                    {
                        ThrowProtocolViolation("LF is expected.");
                    }

                    break;
                }

                if (currentByte == 13)
                {
                    _gotChunck = true;
                    continue;
                }

                if (currentByte == 10)
                {
                    ThrowProtocolViolation("LF is unexpected.");
                }

                if (currentByte == 32) // SP
                {
                    _foundSPCode = true;
                }

                if (!_foundSPCode)
                {
                    _saved.Append((char)currentByte);
                }

                if (_saved.Length > 20)
                {
                    ThrowProtocolViolation("The chunk size is too long.");
                }
            }

            if (!_gotChunck || currentByte != 10)
            {
                return InputChunkState.None;
            }

            _chunkRead = 0;
            try
            {
                _chunkSize = int.Parse(
                  RemoveChunkExtension(_saved.ToString()), NumberStyles.HexNumber);
            }
            catch
            {
                ThrowProtocolViolation("The chunk size cannot be parsed.");
            }

            if (_chunkSize == 0)
            {
                _trailerState = 2;
                return InputChunkState.Trailer;
            }

            return InputChunkState.Data;
        }

        private InputChunkState setTrailer(byte[] buffer, ref int offset, int length)
        {
            // Check if no trailer.
            if (_trailerState == 2 && buffer[offset] == 13 && _saved.Length == 0)
            {
                offset++;
                if (offset < length && buffer[offset] == 10)
                {
                    offset++;
                    return InputChunkState.End;
                }

                offset--;
            }

            while (offset < length && _trailerState < 4)
            {
                var currentByte = buffer[offset++];
                _saved.Append((char)currentByte);
                if (_saved.Length > 4196)
                {
                    ThrowProtocolViolation("The trailer is too long.");
                }

                if (_trailerState == 1 || _trailerState == 3)
                {
                    if (currentByte != 10)
                    {
                        ThrowProtocolViolation("LF is expected.");
                    }

                    _trailerState++;
                    continue;
                }

                if (currentByte == 13)
                {
                    _trailerState++;
                    continue;
                }

                if (currentByte == 10)
                {
                    ThrowProtocolViolation("LF is unexpected.");
                }

                _trailerState = 0;
            }

            if (_trailerState < 4)
            {
                return InputChunkState.Trailer;
            }

            _saved.Length -= 2;
            var reader = new StringReader(_saved.ToString());

            string line;
            while ((line = reader.ReadLine()) != null && line.Length > 0)
            {
                Headers.Add(line);
            }

            return InputChunkState.End;
        }

        private static void ThrowProtocolViolation(string message)
        {
            throw new WebException($"EonaCat Network: {message}", null, WebExceptionStatus.ServerProtocolViolation, null);
        }

        private void Write(byte[] buffer, ref int offset, int length)
        {
            if (_state == InputChunkState.End)
            {
                ThrowProtocolViolation("The chunks were ended.");
            }

            if (_state == InputChunkState.None)
            {
                _state = SetChunkSize(buffer, ref offset, length);
                if (_state == InputChunkState.None)
                {
                    return;
                }

                _saved.Length = 0;
                _gotChunck = false;
                _foundSPCode = false;
            }

            if (_state == InputChunkState.Data && offset < length)
            {
                _state = WriteData(buffer, ref offset, length);
                if (_state == InputChunkState.Data)
                {
                    return;
                }
            }

            if (_state == InputChunkState.DataEnded && offset < length)
            {
                _state = SeekCrLf(buffer, ref offset, length);
                if (_state == InputChunkState.DataEnded)
                {
                    return;
                }

                _gotChunck = false;
            }

            if (_state == InputChunkState.Trailer && offset < length)
            {
                _state = setTrailer(buffer, ref offset, length);
                if (_state == InputChunkState.Trailer)
                {
                    return;
                }

                _saved.Length = 0;
            }

            if (offset < length)
            {
                Write(buffer, ref offset, length);
            }
        }

        private InputChunkState WriteData(byte[] buffer, ref int offset, int length)
        {
            var cnt = length - offset;
            var left = _chunkSize - _chunkRead;
            if (cnt > left)
            {
                cnt = left;
            }

            var data = new byte[cnt];
            Buffer.BlockCopy(buffer, offset, data, 0, cnt);
            _chunks.Add(new WebChunk(data));

            offset += cnt;
            _chunkRead += cnt;

            return _chunkRead == _chunkSize ? InputChunkState.DataEnded : InputChunkState.Data;
        }

        /// <summary>
        /// Resets the internal buffer and state of the chunk stream.
        /// </summary>
        internal void ResetBuffer()
        {
            _chunkRead = 0;
            _chunkSize = -1;
            _chunks.Clear();
        }

        /// <summary>
        /// Writes a specified amount of data to the chunk stream and reads it back.
        /// </summary>
        /// <param name="buffer">The byte array containing data to be written to the chunk stream.</param>
        /// <param name="offset">The offset in the buffer at which to begin writing.</param>
        /// <param name="writeCount">The number of bytes to write to the chunk stream.</param>
        /// <param name="readCount">The number of bytes to read back from the chunk stream.</param>
        /// <returns>The number of bytes read from the chunk stream.</returns>
        internal int WriteAndReadBack(byte[] buffer, int offset, int writeCount, int readCount)
        {
            Write(buffer, offset, writeCount);
            return Read(buffer, offset, readCount);
        }

        /// <summary>
        /// Reads a specified amount of data from the chunk stream.
        /// </summary>
        /// <param name="buffer">The destination buffer.</param>
        /// <param name="offset">The zero-based byte offset in the buffer at which to begin storing the data.</param>
        /// <param name="count">The maximum number of bytes to read.</param>
        /// <returns>The total number of bytes read into the buffer.</returns>
        public int Read(byte[] buffer, int offset, int count)
        {
            if (count <= 0)
            {
                return 0;
            }

            return read(buffer, offset, count);
        }

        /// <summary>
        /// Writes a specified amount of data to the chunk stream.
        /// </summary>
        /// <param name="buffer">The byte array containing data to be written to the chunk stream.</param>
        /// <param name="offset">The offset in the buffer at which to begin writing.</param>
        /// <param name="count">The number of bytes to write to the chunk stream.</param>
        public void Write(byte[] buffer, int offset, int count)
        {
            if (count <= 0)
            {
                return;
            }

            Write(buffer, ref offset, offset + count);
        }
    }
}