// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.IO;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents a stream for handling chunked requests in the EonaCat network library.
    /// </summary>
    internal class ChunkedRequestStream : RequestStream
    {
        private const int _bufferLength = 8192;
        private readonly HttpListenerContext _context;
        private bool _disposed;
        private bool _noMoreData;

        /// <summary>
        /// Initializes a new instance of the <see cref="ChunkedRequestStream"/> class with the specified stream, buffer, offset, count, and context.
        /// </summary>
        /// <param name="stream">The underlying stream.</param>
        /// <param name="buffer">The byte array used for buffering.</param>
        /// <param name="offset">The offset in the buffer at which to begin reading.</param>
        /// <param name="count">The maximum number of bytes to read.</param>
        /// <param name="context">The HTTP listener context.</param>
        internal ChunkedRequestStream(Stream stream, byte[] buffer, int offset, int count, HttpListenerContext context)
          : base(stream, buffer, offset, count)
        {
            _context = context;
            Decoder = new ChunkStream((WebHeaderCollection)context.Request.Headers);
        }

        /// <summary>
        /// Gets or sets the chunk stream decoder associated with the chunked request stream.
        /// </summary>
        internal ChunkStream Decoder { get; set; }

        private void OnRead(IAsyncResult asyncResult)
        {
            var readBufferState = (ReadBufferState)asyncResult.AsyncState;
            var result = readBufferState.AsyncResult;
            try
            {
                var nread = base.EndRead(asyncResult);
                Decoder.Write(result.Buffer, result.Offset, nread);
                nread = Decoder.Read(readBufferState.Buffer, readBufferState.Offset, readBufferState.Count);
                readBufferState.Offset += nread;
                readBufferState.Count -= nread;
                if (readBufferState.Count == 0 || !Decoder.WantMore || nread == 0)
                {
                    _noMoreData = !Decoder.WantMore && nread == 0;
                    result.Count = readBufferState.InitialCount - readBufferState.Count;
                    result.Complete();

                    return;
                }

                result.Offset = 0;
                result.Count = Math.Min(_bufferLength, Decoder.ChunkLeft + 6);
                base.BeginRead(result.Buffer, result.Offset, result.Count, OnRead, readBufferState);
            }
            catch (Exception ex)
            {
                _context.Connection.SendError(ex.Message, 400);
                result.Complete(ex);
            }
        }

        /// <summary>
        /// Begins an asynchronous read operation from the stream.
        /// </summary>
        /// <param name="buffer">The destination buffer.</param>
        /// <param name="offset">The zero-based byte offset in the buffer at which to begin storing the data.</param>
        /// <param name="count">The maximum number of bytes to read.</param>
        /// <param name="callback">An optional asynchronous callback, to be called when the read is complete.</param>
        /// <param name="state">A user-provided object that distinguishes this particular asynchronous read request from other requests.</param>
        /// <returns>An <see cref="IAsyncResult"/> that represents the asynchronous read operation.</returns>
        public override IAsyncResult BeginRead(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
        {
            if (_disposed)
            {
                throw new ObjectDisposedException(GetType().ToString());
            }

            if (buffer == null)
            {
                throw new ArgumentNullException(nameof(buffer));
            }

            if (offset < 0)
            {
                throw new ArgumentOutOfRangeException(nameof(offset), "A negative value.");
            }

            if (count < 0)
            {
                throw new ArgumentOutOfRangeException(nameof(count), "A negative value.");
            }

            var len = buffer.Length;
            if (offset + count > len)
            {
                throw new ArgumentException(
                  "The sum of 'offset' and 'count' is greater than 'buffer' length.");
            }

            var result = new HttpStreamAsyncResult(callback, state);
            if (_noMoreData)
            {
                result.Complete();
                return result;
            }

            var nread = Decoder.Read(buffer, offset, count);
            offset += nread;
            count -= nread;
            if (count == 0)
            {
                result.Count = nread;
                result.Complete();

                return result;
            }

            if (!Decoder.WantMore)
            {
                _noMoreData = nread == 0;
                result.Count = nread;
                result.Complete();

                return result;
            }

            result.Buffer = new byte[_bufferLength];
            result.Offset = 0;
            result.Count = _bufferLength;

            var readBufferState = new ReadBufferState(buffer, offset, count, result);
            readBufferState.InitialCount += nread;
            base.BeginRead(result.Buffer, result.Offset, result.Count, OnRead, readBufferState);

            return result;
        }

        /// <summary>
        /// Closes the stream.
        /// </summary>
        public override void Close()
        {
            if (_disposed)
            {
                return;
            }

            _disposed = true;
            base.Close();
        }

        /// <summary>
        /// Ends an asynchronous read operation from the stream.
        /// </summary>
        /// <param name="asyncResult">The result of the asynchronous operation.</param>
        /// <returns>The number of bytes read from the stream, between zero (0) and the number of bytes you requested. Streams return zero (0) only at the end of the stream, otherwise, they should block until at least one byte is available.</returns>
        public override int EndRead(IAsyncResult asyncResult)
        {
            if (_disposed)
            {
                throw new ObjectDisposedException(GetType().ToString());
            }

            if (asyncResult == null)
            {
                throw new ArgumentNullException(nameof(asyncResult));
            }

            if (asyncResult is not HttpStreamAsyncResult result)
            {
                throw new ArgumentException("A wrong IAsyncResult.", nameof(asyncResult));
            }

            if (!result.IsCompleted)
            {
                result.AsyncWaitHandle.WaitOne();
            }

            if (result.HasException)
            {
                throw new HttpListenerException(400, "I/O operation aborted.");
            }

            return result.Count;
        }

        /// <summary>
        /// Reads a sequence of bytes from the stream and advances the position within the stream by the number of bytes read.
        /// </summary>
        /// <param name="buffer">The byte array to read data into.</param>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin storing the data.</param>
        /// <param name="count">The maximum number of bytes to be read from the stream.</param>
        /// <returns>The total number of bytes read into the buffer. This can be less than the number of bytes requested if that many bytes are not currently available, or zero (0) if the end of the stream has been reached.</returns>
        public override int Read(byte[] buffer, int offset, int count)
        {
            var result = BeginRead(buffer, offset, count, null, null);
            return EndRead(result);
        }
    }
}