// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.IO;
using System.Security.Principal;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents the context of a WebSocket connection within an <see cref="HttpListener"/>.
    /// </summary>
    /// <remarks>
    /// This class provides access to various properties and methods for interacting with the WebSocket connection
    /// within the context of an HTTP request handled by an <see cref="HttpListener"/>.
    /// </remarks>
    /// <seealso cref="WSContext"/>
    public class HttpListenerWSContext : WSContext
    {
        private readonly HttpListenerContext _context;
        private readonly WSClient _websocket;

        /// <summary>
        /// Initializes a new instance of the <see cref="HttpListenerWSContext"/> class.
        /// </summary>
        /// <param name="context">The <see cref="HttpListenerContext"/> associated with the WebSocket connection.</param>
        /// <param name="protocol">The WebSocket protocol negotiated during the connection.</param>
        internal HttpListenerWSContext(HttpListenerContext context, string protocol)
        {
            _context = context;
            _websocket = new WSClient(this, protocol);
        }

        /// <summary>
        /// Gets the stream of the underlying TCP connection.
        /// </summary>
        internal Stream Stream => _context.Connection.Stream;

        public override CookieCollection CookieCollection => _context.Request.Cookies;

        public override NameValueCollection Headers => _context.Request.Headers;

        public override string Host => _context.Request.Headers["Host"];

        public override bool IsAuthenticated => _context.User != null;

        public override bool IsLocal => _context.Request.IsLocal;

        public override bool IsSecureConnection => _context.Connection.IsSecure;

        public override bool IsWebSocketRequest => _context.Request.IsWebSocketRequest;

        public override string Origin => _context.Request.Headers["Origin"];

        public override NameValueCollection QueryString => _context.Request.QueryString;

        public override Uri RequestUri => _context.Request.Url;

        public override string SecWebSocketKey => _context.Request.Headers["Sec-WebSocket-Key"];

        public override IEnumerable<string> SecWebSocketProtocols
        {
            get
            {
                var protocols = _context.Request.Headers["Sec-WebSocket-Protocol"];
                if (protocols != null)
                {
                    foreach (var protocol in protocols.Split(','))
                    {
                        yield return protocol.Trim();
                    }
                }
            }
        }

        public override string SecWebSocketVersion => _context.Request.Headers["Sec-WebSocket-Version"];

        public override System.Net.IPEndPoint ServerEndPoint => _context.Connection.LocalEndPoint;

        public override IPrincipal User => _context.User;

        public override System.Net.IPEndPoint UserEndPoint => _context.Connection.RemoteEndPoint;

        public override WSClient WebSocket => _websocket;

        /// <summary>
        /// Closes the WebSocket connection.
        /// </summary>
        internal void Close()
        {
            _context.Connection.Close(true);
        }

        /// <summary>
        /// Closes the WebSocket connection with the specified HTTP status code.
        /// </summary>
        /// <param name="code">The HTTP status code indicating the reason for closure.</param>
        internal void Close(HttpStatusCode code)
        {
            _context.Response.Close(code);
        }

        /// <inheritdoc />
        public override string ToString()
        {
            return _context.Request.ToString();
        }
    }
}