// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Security.Principal;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents the context of a WebSocket connection.
    /// </summary>
    /// <remarks>
    /// This abstract class defines properties and methods for accessing information related to a WebSocket connection,
    /// such as headers, cookies, authentication status, and more.
    /// </remarks>
    public abstract class WSContext
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="WSContext"/> class.
        /// </summary>
        protected WSContext()
        {
        }

        /// <summary>
        /// Gets the collection of cookies associated with the WebSocket connection.
        /// </summary>
        public abstract CookieCollection CookieCollection { get; }

        /// <summary>
        /// Gets the collection of headers associated with the WebSocket connection.
        /// </summary>
        public abstract NameValueCollection Headers { get; }

        /// <summary>
        /// Gets the host information from the WebSocket connection.
        /// </summary>
        public abstract string Host { get; }

        /// <summary>
        /// Gets a value indicating whether the WebSocket connection is authenticated.
        /// </summary>
        public abstract bool IsAuthenticated { get; }

        /// <summary>
        /// Gets a value indicating whether the WebSocket connection is local.
        /// </summary>
        public abstract bool IsLocal { get; }

        /// <summary>
        /// Gets a value indicating whether the WebSocket connection is secure.
        /// </summary>
        public abstract bool IsSecureConnection { get; }

        /// <summary>
        /// Gets a value indicating whether the request is a WebSocket request.
        /// </summary>
        public abstract bool IsWebSocketRequest { get; }

        /// <summary>
        /// Gets the origin of the WebSocket connection.
        /// </summary>
        public abstract string Origin { get; }

        /// <summary>
        /// Gets the query string information from the WebSocket connection.
        /// </summary>
        public abstract NameValueCollection QueryString { get; }

        /// <summary>
        /// Gets the URI of the WebSocket request.
        /// </summary>
        public abstract Uri RequestUri { get; }

        /// <summary>
        /// Gets the value of the 'Sec-WebSocket-Key' header from the WebSocket connection.
        /// </summary>
        public abstract string SecWebSocketKey { get; }

        /// <summary>
        /// Gets the protocols specified in the 'Sec-WebSocket-Protocol' header from the WebSocket connection.
        /// </summary>
        public abstract IEnumerable<string> SecWebSocketProtocols { get; }

        /// <summary>
        /// Gets the value of the 'Sec-WebSocket-Version' header from the WebSocket connection.
        /// </summary>
        public abstract string SecWebSocketVersion { get; }

        /// <summary>
        /// Gets the local endpoint of the WebSocket server.
        /// </summary>
        public abstract System.Net.IPEndPoint ServerEndPoint { get; }

        /// <summary>
        /// Gets the user associated with the WebSocket connection.
        /// </summary>
        public abstract IPrincipal User { get; }

        /// <summary>
        /// Gets the remote endpoint of the WebSocket user.
        /// </summary>
        public abstract System.Net.IPEndPoint UserEndPoint { get; }

        /// <summary>
        /// Gets the WebSocket instance associated with the context.
        /// </summary>
        public abstract WSClient WebSocket { get; }
    }
}