// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Globalization;
using System.Text;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents an HTTP cookie.
    /// </summary>
    [Serializable]
    public sealed class Cookie
    {
        private string _comment;
        private Uri _commentUri;
        private bool _discard;
        private string _domain;
        private DateTime _expires;
        private bool _httpOnly;
        private string _name;
        private string _path;
        private string _port;
        private int[] _ports;
        private static readonly char[] _reservedCharsForName;
        private static readonly char[] _reservedCharsForValue;
        private bool _secure;
        private readonly DateTime _timestamp;
        private string _value;
        private int _version;

        static Cookie()
        {
            _reservedCharsForName = new[] { ' ', '=', ';', ',', '\n', '\r', '\t' };
            _reservedCharsForValue = new[] { ';', ',' };
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Cookie"/> class.
        /// </summary>
        public Cookie()
        {
            _comment = string.Empty;
            _domain = string.Empty;
            _expires = DateTime.MinValue;
            _name = string.Empty;
            _path = string.Empty;
            _port = string.Empty;
            _ports = new int[0];
            _timestamp = DateTime.Now;
            _value = string.Empty;
            _version = 0;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Cookie"/> class with the specified name and value.
        /// </summary>
        /// <param name="name">The name of the cookie.</param>
        /// <param name="value">The value of the cookie.</param>
        public Cookie(string name, string value)
          : this()
        {
            Name = name;
            Value = value;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Cookie"/> class with the specified name, value, and path.
        /// </summary>
        /// <param name="name">The name of the cookie.</param>
        /// <param name="value">The value of the cookie.</param>
        /// <param name="path">The path for which the cookie is valid.</param>
        public Cookie(string name, string value, string path)
          : this(name, value)
        {
            Path = path;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Cookie"/> class with the specified name, value, path, and domain.
        /// </summary>
        /// <param name="name">The name of the cookie.</param>
        /// <param name="value">The value of the cookie.</param>
        /// <param name="path">The path for which the cookie is valid.</param>
        /// <param name="domain">The domain to which the cookie belongs.</param>
        public Cookie(string name, string value, string path, string domain)
          : this(name, value, path)
        {
            Domain = domain;
        }

        internal bool ExactDomain
        {
            get; set;
        }

        internal int MaxAge
        {
            get
            {
                if (_expires == DateTime.MinValue)
                {
                    return 0;
                }

                var expires = _expires.Kind != DateTimeKind.Local
                              ? _expires.ToLocalTime()
                              : _expires;

                var span = expires - DateTime.Now;
                return span > TimeSpan.Zero
                       ? (int)span.TotalSeconds
                       : 0;
            }
        }

        internal int[] Ports => _ports;

        public string Comment
        {
            get
            {
                return _comment;
            }

            set
            {
                _comment = value ?? string.Empty;
            }
        }

        public Uri CommentUri
        {
            get
            {
                return _commentUri;
            }

            set
            {
                _commentUri = value;
            }
        }

        public bool Discard
        {
            get
            {
                return _discard;
            }

            set
            {
                _discard = value;
            }
        }

        public string Domain
        {
            get
            {
                return _domain;
            }

            set
            {
                if (value.IsNullOrEmpty())
                {
                    _domain = string.Empty;
                    ExactDomain = true;
                }
                else
                {
                    _domain = value;
                    ExactDomain = value[0] != '.';
                }
            }
        }

        public bool Expired
        {
            get
            {
                return _expires != DateTime.MinValue && _expires <= DateTime.Now;
            }

            set
            {
                _expires = value ? DateTime.Now : DateTime.MinValue;
            }
        }

        public DateTime Expires
        {
            get
            {
                return _expires;
            }

            set
            {
                _expires = value;
            }
        }

        public bool HttpOnly
        {
            get
            {
                return _httpOnly;
            }

            set
            {
                _httpOnly = value;
            }
        }

        public string Name
        {
            get
            {
                return _name;
            }

            set
            {
                if (!canSetName(value, out string message))
                {
                    throw new CookieException(message);
                }

                _name = value;
            }
        }

        public string Path
        {
            get
            {
                return _path;
            }

            set
            {
                _path = value ?? string.Empty;
            }
        }

        public string Port
        {
            get
            {
                return _port;
            }

            set
            {
                if (value.IsNullOrEmpty())
                {
                    _port = string.Empty;
                    _ports = new int[0];

                    return;
                }

                if (!value.IsEnclosedIn('"'))
                {
                    throw new CookieException(
                      "The value specified for the Port attribute isn't enclosed in double quotes.");
                }

                if (!tryCreatePorts(value, out _ports, out string err))
                {
                    throw new CookieException(
                      string.Format(
                        "The value specified for the Port attribute contains an invalid value: {0}", err));
                }

                _port = value;
            }
        }

        public bool Secure
        {
            get
            {
                return _secure;
            }

            set
            {
                _secure = value;
            }
        }

        public DateTime TimeStamp => _timestamp;

        public string Value
        {
            get
            {
                return _value;
            }

            set
            {
                if (!canSetValue(value, out string message))
                {
                    throw new CookieException(message);
                }

                _value = value.Length > 0 ? value : "\"\"";
            }
        }

        public int Version
        {
            get
            {
                return _version;
            }

            set
            {
                if (value < 0 || value > 1)
                {
                    throw new ArgumentOutOfRangeException(nameof(value), "Not 0 or 1.");
                }

                _version = value;
            }
        }

        private static bool canSetName(string name, out string message)
        {
            if (name.IsNullOrEmpty())
            {
                message = "The value specified for the Name is null or empty.";
                return false;
            }

            if (name[0] == '$' || name.Contains(_reservedCharsForName))
            {
                message = "The value specified for the Name contains an invalid character.";
                return false;
            }

            message = string.Empty;
            return true;
        }

        private static bool canSetValue(string value, out string message)
        {
            if (value == null)
            {
                message = "The value specified for the Value is null.";
                return false;
            }

            if (value.Contains(_reservedCharsForValue) && !value.IsEnclosedIn('"'))
            {
                message = "The value specified for the Value contains an invalid character.";
                return false;
            }

            message = string.Empty;
            return true;
        }

        private static int hash(int i, int j, int k, int l, int m)
        {
            return i ^
                   (j << 13 | j >> 19) ^
                   (k << 26 | k >> 6) ^
                   (l << 7 | l >> 25) ^
                   (m << 20 | m >> 12);
        }

        private string toResponseStringVersion0()
        {
            var output = new StringBuilder(64);
            output.AppendFormat("{0}={1}", _name, _value);

            if (_expires != DateTime.MinValue)
            {
                output.AppendFormat(
                  "; Expires={0}",
                  _expires.ToUniversalTime().ToString(
                    "ddd, dd'-'MMM'-'yyyy HH':'mm':'ss 'GMT'",
                    CultureInfo.CreateSpecificCulture("en-US")));
            }

            if (!_path.IsNullOrEmpty())
            {
                output.AppendFormat("; Path={0}", _path);
            }

            if (!_domain.IsNullOrEmpty())
            {
                output.AppendFormat("; Domain={0}", _domain);
            }

            if (_secure)
            {
                output.Append("; Secure");
            }

            if (_httpOnly)
            {
                output.Append("; HttpOnly");
            }

            return output.ToString();
        }

        private string toResponseStringVersion1()
        {
            var output = new StringBuilder(64);
            output.AppendFormat("{0}={1}; Version={2}", _name, _value, _version);

            if (_expires != DateTime.MinValue)
            {
                output.AppendFormat("; Max-Age={0}", MaxAge);
            }

            if (!_path.IsNullOrEmpty())
            {
                output.AppendFormat("; Path={0}", _path);
            }

            if (!_domain.IsNullOrEmpty())
            {
                output.AppendFormat("; Domain={0}", _domain);
            }

            if (!_port.IsNullOrEmpty())
            {
                if (_port == "\"\"")
                {
                    output.Append("; Port");
                }
                else
                {
                    output.AppendFormat("; Port={0}", _port);
                }
            }

            if (!_comment.IsNullOrEmpty())
            {
                output.AppendFormat("; Comment={0}", _comment.UrlEncode());
            }

            if (_commentUri != null)
            {
                var url = _commentUri.OriginalString;
                output.AppendFormat("; CommentURL={0}", url.IsToken() ? url : url.Quote());
            }

            if (_discard)
            {
                output.Append("; Discard");
            }

            if (_secure)
            {
                output.Append("; Secure");
            }

            return output.ToString();
        }

        private static bool tryCreatePorts(string value, out int[] result, out string parseError)
        {
            var ports = value.Trim('"').Split(',');
            var len = ports.Length;
            var res = new int[len];
            for (var i = 0; i < len; i++)
            {
                res[i] = int.MinValue;

                var port = ports[i].Trim();
                if (port.Length == 0)
                {
                    continue;
                }

                if (!int.TryParse(port, out res[i]))
                {
                    result = new int[0];
                    parseError = port;

                    return false;
                }
            }

            result = res;
            parseError = string.Empty;

            return true;
        }

        // From client to server
        internal string ToRequestString(Uri uri)
        {
            if (_name.Length == 0)
            {
                return string.Empty;
            }

            if (_version == 0)
            {
                return string.Format("{0}={1}", _name, _value);
            }

            var output = new StringBuilder(64);
            output.AppendFormat("$Version={0}; {1}={2}", _version, _name, _value);

            if (!_path.IsNullOrEmpty())
            {
                output.AppendFormat("; $Path={0}", _path);
            }
            else if (uri != null)
            {
                output.AppendFormat("; $Path={0}", uri.GetAbsolutePath());
            }
            else
            {
                output.Append("; $Path=/");
            }

            var appendDomain = uri == null || uri.Host != _domain;
            if (appendDomain && !_domain.IsNullOrEmpty())
            {
                output.AppendFormat("; $Domain={0}", _domain);
            }

            if (!_port.IsNullOrEmpty())
            {
                if (_port == "\"\"")
                {
                    output.Append("; $Port");
                }
                else
                {
                    output.AppendFormat("; $Port={0}", _port);
                }
            }

            return output.ToString();
        }

        // From server to client
        internal string ToResponseString()
        {
            return _name.Length > 0
                   ? (_version == 0 ? toResponseStringVersion0() : toResponseStringVersion1())
                   : string.Empty;
        }

        /// <inheritdoc />
        public override bool Equals(object comparand)
        {
            return comparand is Cookie cookie &&
                   _name.Equals(cookie.Name, StringComparison.InvariantCultureIgnoreCase) &&
                   _value.Equals(cookie.Value, StringComparison.InvariantCulture) &&
                   _path.Equals(cookie.Path, StringComparison.InvariantCulture) &&
                   _domain.Equals(cookie.Domain, StringComparison.InvariantCultureIgnoreCase) &&
                   _version == cookie.Version;
        }

        /// <inheritdoc />
        public override int GetHashCode()
        {
            return hash(
              StringComparer.InvariantCultureIgnoreCase.GetHashCode(_name),
              _value.GetHashCode(),
              _path.GetHashCode(),
              StringComparer.InvariantCultureIgnoreCase.GetHashCode(_domain),
              _version);
        }

        /// <inheritdoc />
        public override string ToString()
        {
            return ToRequestString(null);
        }
    }
}