// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Text;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents a collection of HTTP cookies.
    /// </summary>
    [Serializable]
    public class CookieCollection : ICollection, IEnumerable
    {
        private readonly List<Cookie> _list;
        private object _sync;

        /// <summary>
        /// Initializes a new instance of the <see cref="CookieCollection"/> class.
        /// </summary>
        public CookieCollection()
        {
            _list = new List<Cookie>();
        }


        internal IList<Cookie> List => _list;

        internal IEnumerable<Cookie> Sorted
        {
            get
            {
                var list = new List<Cookie>(_list);
                if (list.Count > 1)
                {
                    list.Sort(compareCookieWithinSorted);
                }

                return list;
            }
        }

        /// <summary>
        /// Gets the number of cookies in the collection.
        /// </summary>
        public int Count => _list.Count;

        /// <summary>
        /// Gets a value indicating whether the collection is read-only. Always returns true.
        /// </summary>
        public bool IsReadOnly => true;

        /// <summary>
        /// Gets a value indicating whether access to the collection is synchronized (thread-safe). Always returns false.
        /// </summary>
        public bool IsSynchronized => false;

        /// <summary>
        /// Gets or sets the cookie at the specified index.
        /// </summary>
        /// <param name="index">The index of the cookie to get or set.</param>
        /// <returns>The cookie at the specified index.</returns>
        public Cookie this[int index]
        {
            get
            {
                if (index < 0 || index >= _list.Count)
                {
                    throw new ArgumentOutOfRangeException(nameof(index));
                }

                return _list[index];
            }
        }

        /// <summary>
        /// Gets the cookie with the specified name.
        /// </summary>
        /// <param name="name">The name of the cookie to retrieve.</param>
        /// <returns>The cookie with the specified name, or null if the cookie is not found.</returns>
        public Cookie this[string name]
        {
            get
            {
                if (name == null)
                {
                    throw new ArgumentNullException(nameof(name));
                }

                foreach (var cookie in Sorted)
                {
                    if (cookie.Name.Equals(name, StringComparison.InvariantCultureIgnoreCase))
                    {
                        return cookie;
                    }
                }

                return null;
            }
        }

        /// <summary>
        /// Gets an object that can be used to synchronize access to the collection.
        /// </summary>
        public object SyncRoot => _sync ??= ((ICollection)_list).SyncRoot;

        private static int compareCookieWithinSort(Cookie x, Cookie y)
        {
            return (x.Name.Length + x.Value.Length) - (y.Name.Length + y.Value.Length);
        }

        private static int compareCookieWithinSorted(Cookie x, Cookie y)
        {
            var ret = 0;
            return (ret = x.Version - y.Version) != 0
                   ? ret
                   : (ret = x.Name.CompareTo(y.Name)) != 0
                     ? ret
                     : y.Path.Length - x.Path.Length;
        }

        private static CookieCollection parseRequest(string value)
        {
            var cookies = new CookieCollection();

            Cookie cookie = null;
            var ver = 0;
            var pairs = splitCookieHeaderValue(value);
            for (var i = 0; i < pairs.Length; i++)
            {
                var pair = pairs[i].Trim();
                if (pair.Length == 0)
                {
                    continue;
                }

                if (pair.StartsWith("$version", StringComparison.InvariantCultureIgnoreCase))
                {
                    ver = int.Parse(pair.GetValue('=', true));
                }
                else if (pair.StartsWith("$path", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (cookie != null)
                    {
                        cookie.Path = pair.GetValue('=');
                    }
                }
                else if (pair.StartsWith("$domain", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (cookie != null)
                    {
                        cookie.Domain = pair.GetValue('=');
                    }
                }
                else if (pair.StartsWith("$port", StringComparison.InvariantCultureIgnoreCase))
                {
                    var port = pair.Equals("$port", StringComparison.InvariantCultureIgnoreCase)
                               ? "\"\""
                               : pair.GetValue('=');

                    if (cookie != null)
                    {
                        cookie.Port = port;
                    }
                }
                else
                {
                    if (cookie != null)
                    {
                        cookies.Add(cookie);
                    }

                    string name;
                    string val = string.Empty;

                    var pos = pair.IndexOf('=');
                    if (pos == -1)
                    {
                        name = pair;
                    }
                    else if (pos == pair.Length - 1)
                    {
                        name = pair.Substring(0, pos).TrimEnd(' ');
                    }
                    else
                    {
                        name = pair.Substring(0, pos).TrimEnd(' ');
                        val = pair.Substring(pos + 1).TrimStart(' ');
                    }

                    cookie = new Cookie(name, val);
                    if (ver != 0)
                    {
                        cookie.Version = ver;
                    }
                }
            }

            if (cookie != null)
            {
                cookies.Add(cookie);
            }

            return cookies;
        }

        private static CookieCollection parseResponse(string value)
        {
            var cookies = new CookieCollection();

            Cookie cookie = null;
            var pairs = splitCookieHeaderValue(value);
            for (var i = 0; i < pairs.Length; i++)
            {
                var pair = pairs[i].Trim();
                if (pair.Length == 0)
                {
                    continue;
                }

                if (pair.StartsWith("version", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (cookie != null)
                    {
                        cookie.Version = int.Parse(pair.GetValue('=', true));
                    }
                }
                else if (pair.StartsWith("expires", StringComparison.InvariantCultureIgnoreCase))
                {
                    var buff = new StringBuilder(pair.GetValue('='), 32);
                    if (i < pairs.Length - 1)
                    {
                        buff.AppendFormat(", {0}", pairs[++i].Trim());
                    }

                    if (!DateTime.TryParseExact(
                      buff.ToString(),
                      new[] { "ddd, dd'-'MMM'-'yyyy HH':'mm':'ss 'GMT'", "r" },
                      CultureInfo.CreateSpecificCulture("en-US"),
                      DateTimeStyles.AdjustToUniversal | DateTimeStyles.AssumeUniversal,
                      out DateTime expires))
                    {
                        expires = DateTime.Now;
                    }

                    if (cookie != null && cookie.Expires == DateTime.MinValue)
                    {
                        cookie.Expires = expires.ToLocalTime();
                    }
                }
                else if (pair.StartsWith("max-age", StringComparison.InvariantCultureIgnoreCase))
                {
                    var max = int.Parse(pair.GetValue('=', true));
                    var expires = DateTime.Now.AddSeconds(max);
                    if (cookie != null)
                    {
                        cookie.Expires = expires;
                    }
                }
                else if (pair.StartsWith("path", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (cookie != null)
                    {
                        cookie.Path = pair.GetValue('=');
                    }
                }
                else if (pair.StartsWith("domain", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (cookie != null)
                    {
                        cookie.Domain = pair.GetValue('=');
                    }
                }
                else if (pair.StartsWith("port", StringComparison.InvariantCultureIgnoreCase))
                {
                    var port = pair.Equals("port", StringComparison.InvariantCultureIgnoreCase)
                               ? "\"\""
                               : pair.GetValue('=');

                    if (cookie != null)
                    {
                        cookie.Port = port;
                    }
                }
                else if (pair.StartsWith("comment", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (cookie != null)
                    {
                        cookie.Comment = pair.GetValue('=').UrlDecode();
                    }
                }
                else if (pair.StartsWith("commenturl", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (cookie != null)
                    {
                        cookie.CommentUri = pair.GetValue('=', true).ToUri();
                    }
                }
                else if (pair.StartsWith("discard", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (cookie != null)
                    {
                        cookie.Discard = true;
                    }
                }
                else if (pair.StartsWith("secure", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (cookie != null)
                    {
                        cookie.Secure = true;
                    }
                }
                else if (pair.StartsWith("httponly", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (cookie != null)
                    {
                        cookie.HttpOnly = true;
                    }
                }
                else
                {
                    if (cookie != null)
                    {
                        cookies.Add(cookie);
                    }

                    string name;
                    string val = string.Empty;

                    var pos = pair.IndexOf('=');
                    if (pos == -1)
                    {
                        name = pair;
                    }
                    else if (pos == pair.Length - 1)
                    {
                        name = pair.Substring(0, pos).TrimEnd(' ');
                    }
                    else
                    {
                        name = pair.Substring(0, pos).TrimEnd(' ');
                        val = pair.Substring(pos + 1).TrimStart(' ');
                    }

                    cookie = new Cookie(name, val);
                }
            }

            if (cookie != null)
            {
                cookies.Add(cookie);
            }

            return cookies;
        }

        private int searchCookie(Cookie cookie)
        {
            var name = cookie.Name;
            var path = cookie.Path;
            var domain = cookie.Domain;
            var ver = cookie.Version;

            for (var i = _list.Count - 1; i >= 0; i--)
            {
                var c = _list[i];
                if (c.Name.Equals(name, StringComparison.InvariantCultureIgnoreCase) &&
                    c.Path.Equals(path, StringComparison.InvariantCulture) &&
                    c.Domain.Equals(domain, StringComparison.InvariantCultureIgnoreCase) &&
                    c.Version == ver)
                {
                    return i;
                }
            }

            return -1;
        }

        private static string[] splitCookieHeaderValue(string value)
        {
            return new List<string>(value.SplitHeaderValue(',', ';')).ToArray();
        }

        /// <summary>
        /// Parses the specified cookie string, creating a <see cref="CookieCollection"/>.
        /// </summary>
        /// <param name="value">The cookie string to parse.</param>
        /// <param name="response">True if parsing a response header; otherwise, false.</param>
        /// <returns>A <see cref="CookieCollection"/> instance representing the parsed cookies.</returns>
        internal static CookieCollection Parse(string value, bool response)
        {
            return response
                   ? parseResponse(value)
                   : parseRequest(value);
        }

        internal void SetOrRemove(Cookie cookie)
        {
            var pos = searchCookie(cookie);
            if (pos == -1)
            {
                if (!cookie.Expired)
                {
                    _list.Add(cookie);
                }

                return;
            }

            if (!cookie.Expired)
            {
                _list[pos] = cookie;
                return;
            }

            _list.RemoveAt(pos);
        }

        internal void SetOrRemove(CookieCollection cookies)
        {
            foreach (Cookie cookie in cookies)
            {
                SetOrRemove(cookie);
            }
        }

        internal void Sort()
        {
            if (_list.Count > 1)
            {
                _list.Sort(compareCookieWithinSort);
            }
        }

        /// <summary>
        /// Adds the specified cookie to the collection, updating it if it already exists.
        /// </summary>
        /// <param name="cookie">The cookie to add or update.</param>
        public void Add(Cookie cookie)
        {
            if (cookie == null)
            {
                throw new ArgumentNullException(nameof(cookie));
            }

            var pos = searchCookie(cookie);
            if (pos == -1)
            {
                _list.Add(cookie);
                return;
            }

            _list[pos] = cookie;
        }

        /// <summary>
        /// Adds the cookies from the specified <see cref="CookieCollection"/> to this collection, updating existing cookies.
        /// </summary>
        /// <param name="cookies">The <see cref="CookieCollection"/> to add or update from.</param>
        public void Add(CookieCollection cookies)
        {
            if (cookies == null)
            {
                throw new ArgumentNullException(nameof(cookies));
            }

            foreach (Cookie cookie in cookies)
            {
                Add(cookie);
            }
        }

        /// <summary>
        /// Copies the cookies in the collection to the specified array, starting at the specified index.
        /// </summary>
        /// <param name="array">The destination array.</param>
        /// <param name="index">The index in the destination array at which copying begins.</param>
        public void CopyTo(Array array, int index)
        {
            if (array == null)
            {
                throw new ArgumentNullException(nameof(array));
            }

            if (index < 0)
            {
                throw new ArgumentOutOfRangeException(nameof(index), "Less than zero.");
            }

            if (array.Rank > 1)
            {
                throw new ArgumentException("Multidimensional.", nameof(array));
            }

            if (array.Length - index < _list.Count)
            {
                throw new ArgumentException(
                  "The number of elements in this collection is greater than the available space of the destination array.");
            }

            if (!array.GetType().GetElementType().IsAssignableFrom(typeof(Cookie)))
            {
                throw new InvalidCastException(
                  "The elements in this collection cannot be cast automatically to the type of the destination array.");
            } ((IList)_list).CopyTo(array, index);
        }

        /// <summary>
        /// Copies the cookies in the collection to the specified <see cref="Cookie"/> array, starting at the specified index.
        /// </summary>
        /// <param name="array">The destination array.</param>
        /// <param name="index">The index in the destination array at which copying begins.</param>
        public void CopyTo(Cookie[] array, int index)
        {
            if (array == null)
            {
                throw new ArgumentNullException(nameof(array));
            }

            if (index < 0)
            {
                throw new ArgumentOutOfRangeException(nameof(index), "Less than zero.");
            }

            if (array.Length - index < _list.Count)
            {
                throw new ArgumentException(
                  "The number of elements in this collection is greater than the available space of the destination array.");
            }

            _list.CopyTo(array, index);
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>An enumerator for the collection.</returns>
        public IEnumerator GetEnumerator()
        {
            return _list.GetEnumerator();
        }
    }
}