// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents an HTTP connection.
    /// </summary>
    internal sealed class HttpConnection
    {
        private byte[] _buffer;
        private const int _bufferLength = 8192;
        private const int TIMEOUT_CONTINUE = 15000;
        private const int TIMEOUT_INITIAL = 90000;
        private HttpListenerContext _context;
        private bool _contextRegistered;
        private StringBuilder _currentLine;
        private InputState _inputState;
        private RequestStream _inputStream;
        private HttpListener _lastListener;
        private LineState _lineState;
        private readonly EndPointListener _listener;
        private ResponseStream _outputStream;
        private int _position;
        private MemoryStream _requestBuffer;
        private Socket _socket;
        private readonly object _lock;
        private int _timeout;
        private readonly Dictionary<int, bool> _timeoutCanceled;
        private Timer _timer;

        /// <summary>
        /// Initializes a new instance of the <see cref="HttpConnection"/> class.
        /// </summary>
        /// <param name="socket">The socket associated with the connection.</param>
        /// <param name="listener">The endpoint listener.</param>
        internal HttpConnection(Socket socket, EndPointListener listener)
        {
            _socket = socket;
            _listener = listener;
            IsSecure = listener.IsSecure;

            var networkStream = new NetworkStream(socket, false);
            if (IsSecure)
            {
                var conf = listener.SSL;
                var sslStream = new SslStream(networkStream, false, conf.ClientCertificateValidationCallback);
                sslStream.AuthenticateAsServer(
                  conf.Certificate,
                  conf.IsClientCertificateRequired,
                  conf.SslProtocols,
                  conf.CheckForCertificateRevocation
                );

                Stream = sslStream;
            }
            else
            {
                Stream = networkStream;
            }

            _lock = new object();
            _timeout = TIMEOUT_INITIAL;
            _timeoutCanceled = new Dictionary<int, bool>();
            _timer = new Timer(OnTimeout, this, Timeout.Infinite, Timeout.Infinite);

            Setup();
        }

        /// <summary>
        /// Gets a value indicating whether the connection is closed.
        /// </summary>
        public bool IsClosed => _socket == null;

        /// <summary>
        /// Gets a value indicating whether the connection is secure.
        /// </summary>
        public bool IsSecure { get; }

        /// <summary>
        /// Gets the local endpoint.
        /// </summary>
        public IPEndPoint LocalEndPoint => (IPEndPoint)_socket.LocalEndPoint;

        /// <summary>
        /// Gets the remote endpoint.
        /// </summary>
        public IPEndPoint RemoteEndPoint => (IPEndPoint)_socket.RemoteEndPoint;

        /// <summary>
        /// Gets or sets the number of reuses.
        /// </summary>
        public int Reuses { get; private set; }

        /// <summary>
        /// Gets the network stream associated with the connection.
        /// </summary>
        public Stream Stream { get; private set; }

        private void close()
        {
            lock (_lock)
            {
                if (_socket == null)
                {
                    return;
                }

                DisposeTimer();
                DisposeRequestBuffer();
                DisposeStream();
                CloseSocket();
            }

            UnregisterContext();
            RemoveConnection();
        }

        private void CloseSocket()
        {
            try
            {
                _socket.Shutdown(SocketShutdown.Both);
            }
            catch
            {
            }

            _socket.Close();
            _socket = null;
        }

        private void DisposeRequestBuffer()
        {
            if (_requestBuffer == null)
            {
                return;
            }

            _requestBuffer.Dispose();
            _requestBuffer = null;
        }

        private void DisposeStream()
        {
            if (Stream == null)
            {
                return;
            }

            _inputStream = null;
            _outputStream = null;

            Stream.Dispose();
            Stream = null;
        }

        private void DisposeTimer()
        {
            if (_timer == null)
            {
                return;
            }

            try
            {
                _timer.Change(Timeout.Infinite, Timeout.Infinite);
            }
            catch
            {
            }

            _timer.Dispose();
            _timer = null;
        }

        private void Setup()
        {
            _context = new HttpListenerContext(this);
            _inputState = InputState.RequestLine;
            _inputStream = null;
            _lineState = LineState.None;
            _outputStream = null;
            _position = 0;
            _requestBuffer = new MemoryStream();
        }

        private static void OnRead(IAsyncResult asyncResult)
        {
            var httpConnection = (HttpConnection)asyncResult.AsyncState;
            if (httpConnection._socket == null)
            {
                return;
            }

            lock (httpConnection._lock)
            {
                if (httpConnection._socket == null)
                {
                    return;
                }

                var nread = -1;
                var len = 0;
                try
                {
                    var current = httpConnection.Reuses;
                    if (!httpConnection._timeoutCanceled[current])
                    {
                        httpConnection._timer.Change(Timeout.Infinite, Timeout.Infinite);
                        httpConnection._timeoutCanceled[current] = true;
                    }

                    nread = httpConnection.Stream.EndRead(asyncResult);
                    httpConnection._requestBuffer.Write(httpConnection._buffer, 0, nread);
                    len = (int)httpConnection._requestBuffer.Length;
                }
                catch (Exception exception)
                {
                    if (httpConnection._requestBuffer != null && httpConnection._requestBuffer.Length > 0)
                    {
                        httpConnection.SendError(exception.Message, 400);
                        return;
                    }

                    httpConnection.close();
                    return;
                }

                if (nread <= 0)
                {
                    httpConnection.close();
                    return;
                }

                if (httpConnection.ProcessInput(httpConnection._requestBuffer.GetBuffer(), len))
                {
                    if (!httpConnection._context.HasError)
                    {
                        httpConnection._context.Request.FinishInitialization();
                    }

                    if (httpConnection._context.HasError)
                    {
                        httpConnection.SendError();
                        return;
                    }

                    if (!httpConnection._listener.TrySearchHttpListener(httpConnection._context.Request.Url, out HttpListener httpListener))
                    {
                        httpConnection.SendError(null, 404);
                        return;
                    }

                    if (httpConnection._lastListener != httpListener)
                    {
                        httpConnection.RemoveConnection();
                        if (!httpListener.AddConnection(httpConnection))
                        {
                            httpConnection.close();
                            return;
                        }

                        httpConnection._lastListener = httpListener;
                    }

                    httpConnection._context.Listener = httpListener;
                    if (!httpConnection._context.Authenticate())
                    {
                        return;
                    }

                    if (httpConnection._context.Register())
                    {
                        httpConnection._contextRegistered = true;
                    }

                    return;
                }

                httpConnection.Stream.BeginRead(httpConnection._buffer, 0, _bufferLength, OnRead, httpConnection);
            }
        }

        private static void OnTimeout(object state)
        {
            var httpConnection = (HttpConnection)state;
            var current = httpConnection.Reuses;
            if (httpConnection._socket == null)
            {
                return;
            }

            lock (httpConnection._lock)
            {
                if (httpConnection._socket == null)
                {
                    return;
                }

                if (httpConnection._timeoutCanceled[current])
                {
                    return;
                }

                httpConnection.SendError(null, 408);
            }
        }

        private bool ProcessInput(byte[] data, int length)
        {
            _currentLine ??= new StringBuilder(64);

            var nread = 0;
            try
            {
                string line;
                while ((line = ReadLineFrom(data, _position, length, out nread)) != null)
                {
                    _position += nread;
                    if (line.Length == 0)
                    {
                        if (_inputState == InputState.RequestLine)
                        {
                            continue;
                        }

                        if (_position > 32768)
                        {
                            _context.ErrorMessage = "Headers too long";
                        }

                        _currentLine = null;
                        return true;
                    }

                    if (_inputState == InputState.RequestLine)
                    {
                        _context.Request.SetRequestLine(line);
                        _inputState = InputState.Headers;
                    }
                    else
                    {
                        _context.Request.AddHeader(line);
                    }

                    if (_context.HasError)
                    {
                        return true;
                    }
                }
            }
            catch (Exception exception)
            {
                _context.ErrorMessage = exception.Message;
                return true;
            }

            _position += nread;
            if (_position >= 32768)
            {
                _context.ErrorMessage = "Headers too long";
                return true;
            }

            return false;
        }

        private string ReadLineFrom(byte[] buffer, int offset, int length, out int read)
        {
            read = 0;

            for (var i = offset; i < length && _lineState != LineState.LineFeed; i++)
            {
                read++;

                var b = buffer[i];
                if (b == 13)
                {
                    _lineState = LineState.CarriageReturn;
                }
                else if (b == 10)
                {
                    _lineState = LineState.LineFeed;
                }
                else
                {
                    _currentLine.Append((char)b);
                }
            }

            if (_lineState != LineState.LineFeed)
            {
                return null;
            }

            var line = _currentLine.ToString();

            _currentLine.Length = 0;
            _lineState = LineState.None;

            return line;
        }

        private void RemoveConnection()
        {
            if (_lastListener != null)
            {
                _lastListener.RemoveConnection(this);
            }
            else
            {
                _listener.RemoveConnection(this);
            }
        }

        private void UnregisterContext()
        {
            if (!_contextRegistered)
            {
                return;
            }

            _context.Unregister();
            _contextRegistered = false;
        }

        /// <summary>
        /// Closes the connection.
        /// </summary>
        /// <param name="force">True to force close, false otherwise.</param>
        internal void Close(bool force)
        {
            if (_socket == null)
            {
                return;
            }

            lock (_lock)
            {
                if (_socket == null)
                {
                    return;
                }

                if (!force)
                {
                    GetResponseStream().Close(false);
                    if (!_context.Response.CloseConnection && _context.Request.FlushInput())
                    {
                        // Don't close. Keep working.
                        Reuses++;
                        DisposeRequestBuffer();
                        UnregisterContext();
                        Setup();
                        BeginReadRequest();
                        return;
                    }
                }
                else
                {
                    _outputStream?.Close(true);
                }

                close();
            }
        }

        /// <summary>
        /// Initiates reading the request.
        /// </summary>
        public void BeginReadRequest()
        {
            _buffer ??= new byte[_bufferLength];

            if (Reuses == 1)
            {
                _timeout = TIMEOUT_CONTINUE;
            }

            try
            {
                _timeoutCanceled.Add(Reuses, false);
                _timer.Change(_timeout, Timeout.Infinite);
                Stream.BeginRead(_buffer, 0, _bufferLength, OnRead, this);
            }
            catch
            {
                close();
            }
        }

        /// <summary>
        /// Closes the connection.
        /// </summary>
        public void Close()
        {
            Close(false);
        }

        /// <summary>
        /// Gets the request stream.
        /// </summary>
        /// <param name="contentLength">The length of the content.</param>
        /// <param name="chunked">True if chunked, false otherwise.</param>
        /// <returns>The request stream.</returns>
        public RequestStream GetRequestStream(long contentLength, bool chunked)
        {
            if (_inputStream != null || _socket == null)
            {
                return _inputStream;
            }

            lock (_lock)
            {
                if (_socket == null)
                {
                    return _inputStream;
                }

                var buff = _requestBuffer.GetBuffer();
                var len = (int)_requestBuffer.Length;
                DisposeRequestBuffer();
                if (chunked)
                {
                    _context.Response.SendInChunks = true;
                    _inputStream =
                      new ChunkedRequestStream(Stream, buff, _position, len - _position, _context);
                }
                else
                {
                    _inputStream =
                      new RequestStream(Stream, buff, _position, len - _position, contentLength);
                }

                return _inputStream;
            }
        }

        /// <summary>
        /// Gets the response stream.
        /// </summary>
        /// <returns>The response stream.</returns>
        public ResponseStream GetResponseStream()
        {
            if (_outputStream != null || _socket == null)
            {
                return _outputStream;
            }

            lock (_lock)
            {
                if (_socket == null)
                {
                    return _outputStream;
                }

                var lsnr = _context.Listener;
                var ignore = lsnr == null || lsnr.IgnoreWriteExceptions;
                _outputStream = new ResponseStream(Stream, _context.Response, ignore);

                return _outputStream;
            }
        }

        /// <summary>
        /// Sends an error response.
        /// </summary>
        public void SendError()
        {
            SendError(_context.ErrorMessage, _context.ErrorStatus);
        }

        /// <summary>
        /// Sends an error response with the specified message and status code.
        /// </summary>
        /// <param name="message">The error message.</param>
        /// <param name="status">The HTTP status code.</param>
        public void SendError(string message, int status)
        {
            if (_socket == null)
            {
                return;
            }

            lock (_lock)
            {
                if (_socket == null)
                {
                    return;
                }

                try
                {
                    var httpResponse = _context.Response;
                    httpResponse.StatusCode = status;
                    httpResponse.ContentType = "text/html";

                    var content = new StringBuilder(64);
                    content.AppendFormat("<html><head><title>EonaCat.Network Error</title></head><body><h1>{0} {1}", status, httpResponse.StatusDescription);
                    if (message != null && message.Length > 0)
                    {
                        content.AppendFormat(" ({0})</h1></body></html>", message);
                    }
                    else
                    {
                        content.Append("</h1></body></html>");
                    }

                    var encoding = Encoding.UTF8;
                    var entity = encoding.GetBytes(content.ToString());
                    httpResponse.ContentEncoding = encoding;
                    httpResponse.ContentLength64 = entity.LongLength;

                    httpResponse.Close(entity, true);
                }
                catch
                {
                    Close(true);
                }
            }
        }
    }
}