// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System.Collections.Specialized;
using System.Security.Principal;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents an HTTP Digest identity.
    /// </summary>
    public class HttpDigestIdentity : GenericIdentity
    {
        private readonly NameValueCollection _parameters;

        /// <summary>
        /// Initializes a new instance of the <see cref="HttpDigestIdentity"/> class.
        /// </summary>
        /// <param name="parameters">The collection of parameters.</param>
        internal HttpDigestIdentity(NameValueCollection parameters)
          : base(parameters["username"], "Digest")
        {
            _parameters = parameters;
        }

        /// <summary>
        /// Gets the algorithm used for digest authentication.
        /// </summary>
        public string Algorithm => _parameters["algorithm"];

        /// <summary>
        /// Gets the client nonce.
        /// </summary>
        public string Cnonce => _parameters["cnonce"];

        /// <summary>
        /// Gets the nonce count.
        /// </summary>
        public string Nc => _parameters["nc"];

        /// <summary>
        /// Gets the nonce value.
        /// </summary>
        public string Nonce => _parameters["nonce"];

        /// <summary>
        /// Gets the opaque value.
        /// </summary>
        public string Opaque => _parameters["opaque"];

        /// <summary>
        /// Gets the quality of protection.
        /// </summary>
        public string Qop => _parameters["qop"];

        /// <summary>
        /// Gets the realm.
        /// </summary>
        public string Realm => _parameters["realm"];

        /// <summary>
        /// Gets the response value.
        /// </summary>
        public string Response => _parameters["response"];

        /// <summary>
        /// Gets the URI.
        /// </summary>
        public string Uri => _parameters["uri"];

        /// <summary>
        /// Validates the HTTP Digest identity.
        /// </summary>
        /// <param name="password">The password.</param>
        /// <param name="realm">The realm.</param>
        /// <param name="method">The HTTP method.</param>
        /// <param name="entity">The entity.</param>
        /// <returns>True if the identity is valid, false otherwise.</returns>
        internal bool IsValid(
          string password, string realm, string method, string entity
        )
        {
            var copied = new NameValueCollection(_parameters)
            {
                ["password"] = password,
                ["realm"] = realm,
                ["method"] = method,
                ["entity"] = entity
            };

            var expected = AuthenticationResponse.CreateRequestDigest(copied);
            return _parameters["response"] == expected;
        }
    }
}