// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Security.Principal;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents an HTTP listener that listens for incoming requests.
    /// </summary>
    public sealed class HttpListener : IDisposable
    {
        private AuthenticationSchemes _authSchemes;
        private Func<HttpListenerRequest, AuthenticationSchemes> _authSchemeSelector;
        private string _certFolderPath;
        private readonly Dictionary<HttpConnection, HttpConnection> _connections;
        private readonly object _connectionsSync;
        private readonly List<HttpListenerContext> contextQueue;
        private readonly object _contextQueueLock;
        private readonly Dictionary<HttpListenerContext, HttpListenerContext> _contextRegistry;
        private readonly object _contextRegistryLock;
        private static readonly string _defaultRealm;
        private bool _ignoreWriteExceptions;
        private volatile bool _listening;
        private readonly HttpListenerPrefixCollection _prefixes;
        private string _realm;
        private SSLConfigServer _sslConfig;
        private Func<IIdentity, NetworkCredential> _userCredFinder;
        private readonly List<HttpListenerAsyncResult> _waitQueue;
        private readonly object _waitQueueLock;

        static HttpListener()
        {
            _defaultRealm = "SECRET AREA";
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="HttpListener"/> class.
        /// </summary>
        public HttpListener()
        {
            _authSchemes = AuthenticationSchemes.Anonymous;

            _connections = new Dictionary<HttpConnection, HttpConnection>();
            _connectionsSync = ((ICollection)_connections).SyncRoot;

            contextQueue = new List<HttpListenerContext>();
            _contextQueueLock = ((ICollection)contextQueue).SyncRoot;

            _contextRegistry = new Dictionary<HttpListenerContext, HttpListenerContext>();
            _contextRegistryLock = ((ICollection)_contextRegistry).SyncRoot;

            _prefixes = new HttpListenerPrefixCollection(this);

            _waitQueue = new List<HttpListenerAsyncResult>();
            _waitQueueLock = ((ICollection)_waitQueue).SyncRoot;
        }

        internal bool IsDisposed { get; private set; }

        internal bool ReuseAddress { get; set; }

        /// <summary>
        /// Gets or sets the authentication schemes used by this listener.
        /// </summary>
        public AuthenticationSchemes AuthenticationSchemes
        {
            get
            {
                CheckDisposed();
                return _authSchemes;
            }

            set
            {
                CheckDisposed();
                _authSchemes = value;
            }
        }

        public Func<HttpListenerRequest, AuthenticationSchemes> AuthenticationSchemeSelector
        {
            get
            {
                CheckDisposed();
                return _authSchemeSelector;
            }

            set
            {
                CheckDisposed();
                _authSchemeSelector = value;
            }
        }

        public string CertificateFolderPath
        {
            get
            {
                CheckDisposed();
                return _certFolderPath;
            }

            set
            {
                CheckDisposed();
                _certFolderPath = value;
            }
        }

        public bool IgnoreWriteExceptions
        {
            get
            {
                CheckDisposed();
                return _ignoreWriteExceptions;
            }

            set
            {
                CheckDisposed();
                _ignoreWriteExceptions = value;
            }
        }

        public bool IsListening => _listening;

        public static bool IsSupported => true;

        public HttpListenerPrefixCollection Prefixes
        {
            get
            {
                CheckDisposed();
                return _prefixes;
            }
        }

        public string Realm
        {
            get
            {
                CheckDisposed();
                return _realm;
            }

            set
            {
                CheckDisposed();
                _realm = value;
            }
        }

        public SSLConfigServer SslConfiguration
        {
            get
            {
                CheckDisposed();
                return _sslConfig ??= new SSLConfigServer();
            }

            set
            {
                CheckDisposed();
                _sslConfig = value;
            }
        }

        public bool UnsafeConnectionNtlmAuthentication
        {
            get
            {
                throw new NotSupportedException();
            }

            set
            {
                throw new NotSupportedException();
            }
        }

        public Func<IIdentity, NetworkCredential> UserCredentialsFinder
        {
            get
            {
                CheckDisposed();
                return _userCredFinder;
            }

            set
            {
                CheckDisposed();
                _userCredFinder = value;
            }
        }

        private void CleanupConnections()
        {
            HttpConnection[] httpConnections = null;
            lock (_connectionsSync)
            {
                if (_connections.Count == 0)
                {
                    return;
                }

                // Need to copy this since closing will call the RemoveConnection method.
                var keys = _connections.Keys;
                httpConnections = new HttpConnection[keys.Count];
                keys.CopyTo(httpConnections, 0);
                _connections.Clear();
            }

            for (var i = httpConnections.Length - 1; i >= 0; i--)
            {
                httpConnections[i].Close(true);
            }
        }

        private void CleanupContextQueue(bool sendServiceUnavailable)
        {
            HttpListenerContext[] httpContextQueues = null;
            lock (_contextQueueLock)
            {
                if (contextQueue.Count == 0)
                {
                    return;
                }

                httpContextQueues = contextQueue.ToArray();
                contextQueue.Clear();
            }

            if (!sendServiceUnavailable)
            {
                return;
            }

            foreach (var currentContext in httpContextQueues)
            {
                var response = currentContext.Response;
                response.StatusCode = (int)HttpStatusCode.ServiceUnavailable;
                response.Close();
            }
        }

        private void CleanupContextRegistry()
        {
            HttpListenerContext[] contexts = null;
            lock (_contextRegistryLock)
            {
                if (_contextRegistry.Count == 0)
                {
                    return;
                }

                // Need to copy this since closing will call the UnregisterContext method.
                var keys = _contextRegistry.Keys;
                contexts = new HttpListenerContext[keys.Count];
                keys.CopyTo(contexts, 0);
                _contextRegistry.Clear();
            }

            for (var i = contexts.Length - 1; i >= 0; i--)
            {
                contexts[i].Connection.Close(true);
            }
        }

        private void CleanupWaitQueue(Exception exception)
        {
            HttpListenerAsyncResult[] results = null;
            lock (_waitQueueLock)
            {
                if (_waitQueue.Count == 0)
                {
                    return;
                }

                results = _waitQueue.ToArray();
                _waitQueue.Clear();
            }

            foreach (var currentResult in results)
            {
                currentResult.Complete(exception);
            }
        }

        private void close(bool force)
        {
            if (_listening)
            {
                _listening = false;
                EndPointManager.RemoveListener(this);
            }

            lock (_contextRegistryLock)
            {
                CleanupContextQueue(!force);
            }

            CleanupContextRegistry();
            CleanupConnections();
            CleanupWaitQueue(new ObjectDisposedException(GetType().ToString()));

            IsDisposed = true;
        }

        private HttpListenerAsyncResult GetAsyncResultFromQueue()
        {
            if (_waitQueue.Count == 0)
            {
                return null;
            }

            var ares = _waitQueue[0];
            _waitQueue.RemoveAt(0);

            return ares;
        }

        private HttpListenerContext GetContextFromQueue()
        {
            if (contextQueue.Count == 0)
            {
                return null;
            }

            var ctx = contextQueue[0];
            contextQueue.RemoveAt(0);

            return ctx;
        }

        internal bool AddConnection(HttpConnection connection)
        {
            if (!_listening)
            {
                return false;
            }

            lock (_connectionsSync)
            {
                if (!_listening)
                {
                    return false;
                }

                _connections[connection] = connection;
                return true;
            }
        }

        internal HttpListenerAsyncResult BeginGetContext(HttpListenerAsyncResult asyncResult)
        {
            lock (_contextRegistryLock)
            {
                if (!_listening)
                {
                    throw new HttpListenerException(995);
                }

                var ctx = GetContextFromQueue();
                if (ctx == null)
                {
                    _waitQueue.Add(asyncResult);
                }
                else
                {
                    asyncResult.Complete(ctx, true);
                }

                return asyncResult;
            }
        }

        internal void CheckDisposed()
        {
            if (IsDisposed)
            {
                throw new ObjectDisposedException(GetType().ToString());
            }
        }

        internal string GetRealm()
        {
            var realm = _realm;
            return realm != null && realm.Length > 0 ? realm : _defaultRealm;
        }

        internal Func<IIdentity, NetworkCredential> GetUserCredentialsFinder()
        {
            return _userCredFinder;
        }

        internal bool RegisterContext(HttpListenerContext context)
        {
            if (!_listening)
            {
                return false;
            }

            lock (_contextRegistryLock)
            {
                if (!_listening)
                {
                    return false;
                }

                _contextRegistry[context] = context;

                var result = GetAsyncResultFromQueue();
                if (result == null)
                {
                    contextQueue.Add(context);
                }
                else
                {
                    result.Complete(context);
                }

                return true;
            }
        }

        internal void RemoveConnection(HttpConnection connection)
        {
            lock (_connectionsSync)
            {
                _connections.Remove(connection);
            }
        }

        internal AuthenticationSchemes SelectAuthenticationScheme(HttpListenerRequest request)
        {
            var selector = _authSchemeSelector;
            if (selector == null)
            {
                return _authSchemes;
            }

            try
            {
                return selector(request);
            }
            catch
            {
                return AuthenticationSchemes.None;
            }
        }

        internal void UnregisterContext(HttpListenerContext context)
        {
            lock (_contextRegistryLock)
            {
                _contextRegistry.Remove(context);
            }
        }

        /// <summary>
        /// Aborts the listener and releases all resources associated with it.
        /// </summary>
        public void Abort()
        {
            if (IsDisposed)
            {
                return;
            }

            close(true);
        }

        /// <summary>
        /// Begins asynchronously getting an HTTP context from the listener.
        /// </summary>
        /// <param name="callback">The method to call when the operation completes.</param>
        /// <param name="state">A user-defined object that contains information about the asynchronous operation.</param>
        /// <returns>An <see cref="IAsyncResult"/> that represents the asynchronous operation.</returns>
        public IAsyncResult BeginGetContext(AsyncCallback callback, object state)
        {
            CheckDisposed();
            if (_prefixes.Count == 0)
            {
                throw new InvalidOperationException("The listener has no URI prefix on which listens.");
            }

            if (!_listening)
            {
                throw new InvalidOperationException("The listener hasn't been started.");
            }

            return BeginGetContext(new HttpListenerAsyncResult(callback, state));
        }

        /// <summary>
        /// Closes the listener.
        /// </summary>
        public void Close()
        {
            if (IsDisposed)
            {
                return;
            }

            close(false);
        }

        /// <summary>
        /// Ends an asynchronous operation to get an HTTP context from the listener.
        /// </summary>
        /// <param name="asyncResult">The reference to the pending asynchronous request to finish.</param>
        /// <returns>An <see cref="HttpListenerContext"/> that represents the context of the asynchronous operation.</returns>
        public HttpListenerContext EndGetContext(IAsyncResult asyncResult)
        {
            CheckDisposed();
            if (asyncResult == null)
            {
                throw new ArgumentNullException(nameof(asyncResult));
            }

            if (asyncResult is not HttpListenerAsyncResult result)
            {
                throw new ArgumentException("A wrong IAsyncResult.", nameof(asyncResult));
            }

            if (result.EndCalled)
            {
                throw new InvalidOperationException("This IAsyncResult cannot be reused.");
            }

            result.EndCalled = true;
            if (!result.IsCompleted)
            {
                result.AsyncWaitHandle.WaitOne();
            }

            return result.GetContext();
        }

        /// <summary>
        /// Gets the next available HTTP context from the listener.
        /// </summary>
        /// <returns>An <see cref="HttpListenerContext"/> that represents the context of the HTTP request.</returns>
        public HttpListenerContext GetContext()
        {
            CheckDisposed();
            if (_prefixes.Count == 0)
            {
                throw new InvalidOperationException("The listener has no URI prefix on which listens.");
            }

            if (!_listening)
            {
                throw new InvalidOperationException("The listener hasn't been started.");
            }

            var result = BeginGetContext(new HttpListenerAsyncResult(null, null));
            result.InGet = true;

            return EndGetContext(result);
        }

        /// <summary>
        /// Starts listening for incoming requests.
        /// </summary>
        public void Start()
        {
            CheckDisposed();
            if (_listening)
            {
                return;
            }

            EndPointManager.AddListener(this);
            _listening = true;
        }

        /// <summary>
        /// Stops listening for incoming requests.
        /// </summary>
        public void Stop()
        {
            CheckDisposed();
            if (!_listening)
            {
                return;
            }

            _listening = false;
            EndPointManager.RemoveListener(this);

            lock (_contextRegistryLock)
            {
                CleanupContextQueue(true);
            }

            CleanupContextRegistry();
            CleanupConnections();
            CleanupWaitQueue(new HttpListenerException(995, "The listener is closed."));
        }

        /// <summary>
        /// Disposes of the resources used by the <see cref="HttpListener"/>.
        /// </summary>
        void IDisposable.Dispose()
        {
            if (IsDisposed)
            {
                return;
            }

            close(true);
        }
    }
}