// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Security.Principal;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents the context for an HTTP listener.
    /// </summary>
    public sealed class HttpListenerContext
    {
        private HttpListenerWSContext _websocketContext;

        /// <summary>
        /// Initializes a new instance of the <see cref="HttpListenerContext"/> class.
        /// </summary>
        /// <param name="connection">The underlying <see cref="HttpConnection"/> for the context.</param>
        internal HttpListenerContext(HttpConnection connection)
        {
            Connection = connection;
            ErrorStatus = 400;
            Request = new HttpListenerRequest(this);
            Response = new HttpListenerResponse(this);
        }

        /// <summary>
        /// Gets the underlying <see cref="HttpConnection"/> for the context.
        /// </summary>
        internal HttpConnection Connection { get; }

        /// <summary>
        /// Gets or sets the error message associated with the context.
        /// </summary>
        internal string ErrorMessage { get; set; }

        /// <summary>
        /// Gets or sets the error status associated with the context.
        /// </summary>
        internal int ErrorStatus { get; set; }

        /// <summary>
        /// Gets a value indicating whether an error has occurred.
        /// </summary>
        internal bool HasError => ErrorMessage != null;

        /// <summary>
        /// Gets or sets the <see cref="HttpListener"/> associated with the context.
        /// </summary>
        internal HttpListener Listener { get; set; }

        /// <summary>
        /// Gets the <see cref="HttpListenerRequest"/> associated with the context.
        /// </summary>
        public HttpListenerRequest Request { get; }

        /// <summary>
        /// Gets the <see cref="HttpListenerResponse"/> associated with the context.
        /// </summary>
        public HttpListenerResponse Response { get; }

        /// <summary>
        /// Gets or sets the <see cref="IPrincipal"/> associated with the user.
        /// </summary>
        public IPrincipal User { get; private set; }

        /// <summary>
        /// Authenticates the user based on the specified authentication scheme.
        /// </summary>
        /// <returns><c>true</c> if authentication is successful; otherwise, <c>false</c>.</returns>
        internal bool Authenticate()
        {
            var schm = Listener.SelectAuthenticationScheme(Request);
            if (schm == AuthenticationSchemes.Anonymous)
            {
                return true;
            }

            if (schm == AuthenticationSchemes.None)
            {
                Response.Close(HttpStatusCode.Forbidden);
                return false;
            }

            var realm = Listener.GetRealm();
            var user =
              HttpUtility.CreateUser(
                Request.Headers["Authorization"],
                schm,
                realm,
                Request.HttpMethod,
                Listener.GetUserCredentialsFinder()
              );

            if (user == null || !user.Identity.IsAuthenticated)
            {
                Response.CloseWithAuthChallenge(new AuthenticationChallenge(schm, realm).ToString());
                return false;
            }

            User = user;
            return true;
        }

        /// <summary>
        /// Registers the context with the associated <see cref="HttpListener"/>.
        /// </summary>
        /// <returns><c>true</c> if registration is successful; otherwise, <c>false</c>.</returns>
        internal bool Register()
        {
            return Listener.RegisterContext(this);
        }

        /// <summary>
        /// Unregisters the context from the associated <see cref="HttpListener"/>.
        /// </summary>
        internal void Unregister()
        {
            Listener.UnregisterContext(this);
        }

        /// <summary>
        /// Accepts a WebSocket connection with the specified protocol.
        /// </summary>
        /// <param name="protocol">The WebSocket subprotocol to negotiate.</param>
        /// <returns>The <see cref="HttpListenerWSContext"/> for the WebSocket connection.</returns>
        public HttpListenerWSContext AcceptWebSocket(string protocol)
        {
            if (_websocketContext != null)
            {
                throw new InvalidOperationException("Accepting already in progress.");
            }

            if (protocol != null)
            {
                if (protocol.Length == 0)
                {
                    throw new ArgumentException("Empty string.", nameof(protocol));
                }

                if (!protocol.IsToken())
                {
                    throw new ArgumentException("Contains invalid characters", nameof(protocol));
                }
            }

            _websocketContext = new HttpListenerWSContext(this, protocol);
            return _websocketContext;
        }
    }
}