// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Globalization;
using System.IO;
using System.Security.Cryptography.X509Certificates;
using System.Text;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents an HTTP listener request.
    /// </summary>
    public sealed class HttpListenerRequest
    {
        private static readonly byte[] _100continue;
        private bool _chunked;
        private Encoding _contentEncoding;
        private bool _contentLengthSet;
        private readonly HttpListenerContext _context;
        private CookieCollection _cookies;
        private readonly WebHeaderCollection _headers;
        private Stream _inputStream;
        private bool _keepAlive;
        private bool _keepAliveSet;
        private NameValueCollection _queryString;
        private string _uri;
        private Version _version;
        private bool _websocketRequest;
        private bool _websocketRequestSet;

        static HttpListenerRequest()
        {
            _100continue = Encoding.ASCII.GetBytes("HTTP/1.1 100 Continue\r\n\r\n");
        }

        internal HttpListenerRequest(HttpListenerContext context)
        {
            _context = context;
            ContentLength64 = -1;
            _headers = new WebHeaderCollection();
            RequestTraceIdentifier = Guid.NewGuid();
        }

        /// <summary>
        /// Gets the accepted types for the request.
        /// </summary>
        public string[] AcceptTypes { get; private set; }

        /// <summary>
        /// Gets the content encoding of the request.
        /// </summary>
        public Encoding ContentEncoding => _contentEncoding ??= Encoding.Default;

        /// <summary>
        /// Gets the content length of the request.
        /// </summary>
        public long ContentLength64 { get; private set; }

        /// <summary>
        /// Gets the content type of the request.
        /// </summary>
        public string ContentType => _headers["Content-Type"];

        /// <summary>
        /// Gets the collection of cookies in the request.
        /// </summary>
        public CookieCollection Cookies => _cookies ??= _headers.GetCookies(false);

        /// <summary>
        /// Gets a value indicating whether the request has an entity body.
        /// </summary>
        public bool HasEntityBody => ContentLength64 > 0 || _chunked;

        /// <summary>
        /// Gets the collection of headers in the request.
        /// </summary>
        public NameValueCollection Headers => _headers;

        /// <summary>
        /// Gets the HTTP method of the request.
        /// </summary>
        public string HttpMethod { get; private set; }

        /// <summary>
        /// Gets the input stream of the request.
        /// </summary>
        public Stream InputStream => _inputStream ??= HasEntityBody
                                       ? _context.Connection.GetRequestStream(ContentLength64, _chunked)
                                       : Stream.Null;

        /// <summary>
        /// Gets a value indicating whether the user is authenticated.
        /// </summary>
        public bool IsAuthenticated => _context.User != null;

        /// <summary>
        /// Gets a value indicating whether the request is from a local address.
        /// </summary>
        public bool IsLocal => RemoteEndPoint.Address.IsLocal();

        /// <summary>
        /// Gets a value indicating whether the connection is secure.
        /// </summary>
        public bool IsSecureConnection => _context.Connection.IsSecure;

        /// <summary>
        /// Gets a value indicating whether the request is a WebSocket request.
        /// </summary>
        public bool IsWebSocketRequest
        {
            get
            {
                if (!_websocketRequestSet)
                {
                    _websocketRequest = HttpMethod == "GET" &&
                                        _version > HttpVersion.Version10 &&
                                        _headers.Contains("Upgrade", "websocket") &&
                                        _headers.Contains("Connection", "Upgrade");

                    _websocketRequestSet = true;
                }

                return _websocketRequest;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the connection should be kept alive.
        /// </summary>
        public bool KeepAlive
        {
            get
            {
                if (!_keepAliveSet)
                {
                    string keepAlive;
                    _keepAlive = _version > HttpVersion.Version10 ||
                                 _headers.Contains("Connection", "keep-alive") ||
                                 ((keepAlive = _headers["Keep-Alive"]) != null && keepAlive != "closed");

                    _keepAliveSet = true;
                }

                return _keepAlive;
            }
        }

        /// <summary>
        /// Gets the local endpoint of the connection.
        /// </summary>
        public System.Net.IPEndPoint LocalEndPoint => _context.Connection.LocalEndPoint;

        /// <summary>
        /// Gets the protocol version of the request.
        /// </summary>
        public Version ProtocolVersion => _version;

        /// <summary>
        /// Gets the query string in the request.
        /// </summary>
        public NameValueCollection QueryString => _queryString ??= HttpUtility.InternalParseQueryString(Url.Query, Encoding.UTF8);

        /// <summary>
        /// Gets the raw URL of the request.
        /// </summary>
        public string RawUrl => Url.PathAndQuery;

        /// <summary>
        /// Gets the remote endpoint of the connection.
        /// </summary>
        public System.Net.IPEndPoint RemoteEndPoint => _context.Connection.RemoteEndPoint;

        /// <summary>
        /// Gets the trace identifier for the request.
        /// </summary>
        public Guid RequestTraceIdentifier { get; }

        /// <summary>
        /// Gets the URL of the request.
        /// </summary>
        public Uri Url { get; private set; }

        /// <summary>
        /// Gets the URL referrer of the request.
        /// </summary>
        public Uri UrlReferrer { get; private set; }

        /// <summary>
        /// Gets the user agent of the request.
        /// </summary>
        public string UserAgent => _headers["User-Agent"];

        /// <summary>
        /// Gets the user host address of the request.
        /// </summary>
        public string UserHostAddress => LocalEndPoint.ToString();

        /// <summary>
        /// Gets the user host name of the request.
        /// </summary>
        public string UserHostName => _headers["Host"];

        /// <summary>
        /// Gets the user languages of the request.
        /// </summary>
        public string[] UserLanguages { get; private set; }

        private static bool tryCreateVersion(string version, out Version result)
        {
            try
            {
                result = new Version(version);
                return true;
            }
            catch
            {
                result = null;
                return false;
            }
        }

        internal void AddHeader(string header)
        {
            var colon = header.IndexOf(':');
            if (colon == -1)
            {
                _context.ErrorMessage = "Invalid header";
                return;
            }

            var name = header.Substring(0, colon).Trim();
            var val = header.Substring(colon + 1).Trim();
            _headers.InternalSet(name, val, false);

            var lower = name.ToLower(CultureInfo.InvariantCulture);
            if (lower == "accept")
            {
                AcceptTypes = new List<string>(val.SplitHeaderValue(',')).ToArray();
                return;
            }

            if (lower == "accept-language")
            {
                UserLanguages = val.Split(',');
                return;
            }

            if (lower == "content-length")
            {
                if (long.TryParse(val, out long len) && len >= 0)
                {
                    ContentLength64 = len;
                    _contentLengthSet = true;
                }
                else
                {
                    _context.ErrorMessage = "Invalid Content-Length header";
                }

                return;
            }

            if (lower == "content-type")
            {
                try
                {
                    _contentEncoding = HttpUtility.GetEncoding(val);
                }
                catch
                {
                    _context.ErrorMessage = "Invalid Content-Type header";
                }

                return;
            }

            if (lower == "referer")
            {
                UrlReferrer = val.ToUri();
            }
        }

        internal void FinishInitialization()
        {
            var host = _headers["Host"];
            var nohost = host == null || host.Length == 0;
            if (_version > HttpVersion.Version10 && nohost)
            {
                _context.ErrorMessage = "Invalid Host header";
                return;
            }

            if (nohost)
            {
                host = UserHostAddress;
            }

            Url = HttpUtility.CreateRequestUrl(_uri, host, IsWebSocketRequest, IsSecureConnection);
            if (Url == null)
            {
                _context.ErrorMessage = "Invalid request url";
                return;
            }

            var enc = Headers["Transfer-Encoding"];
            if (_version > HttpVersion.Version10 && enc != null && enc.Length > 0)
            {
                _chunked = enc.ToLower() == "chunked";
                if (!_chunked)
                {
                    _context.ErrorMessage = string.Empty;
                    _context.ErrorStatus = 501;

                    return;
                }
            }

            if (!_chunked && !_contentLengthSet)
            {
                var method = HttpMethod.ToLower();
                if (method == "post" || method == "put")
                {
                    _context.ErrorMessage = string.Empty;
                    _context.ErrorStatus = 411;

                    return;
                }
            }

            var expect = Headers["Expect"];
            if (expect != null && expect.Length > 0 && expect.ToLower() == "100-continue")
            {
                var output = _context.Connection.GetResponseStream();
                output.InternalWrite(_100continue, 0, _100continue.Length);
            }
        }

        // Returns true is the stream could be reused.
        internal bool FlushInput()
        {
            if (!HasEntityBody)
            {
                return true;
            }

            var len = 2048;
            if (ContentLength64 > 0)
            {
                len = (int)Math.Min(ContentLength64, len);
            }

            var buff = new byte[len];
            while (true)
            {
                try
                {
                    var result = InputStream.BeginRead(buff, 0, len, null, null);
                    if (!result.IsCompleted && !result.AsyncWaitHandle.WaitOne(100))
                    {
                        return false;
                    }

                    if (InputStream.EndRead(result) <= 0)
                    {
                        return true;
                    }
                }
                catch
                {
                    return false;
                }
            }
        }

        internal void SetRequestLine(string requestLine)
        {
            var parts = requestLine.Split(new[] { ' ' }, 3);
            if (parts.Length != 3)
            {
                _context.ErrorMessage = "Invalid request line (parts)";
                return;
            }

            HttpMethod = parts[0];
            if (!HttpMethod.IsToken())
            {
                _context.ErrorMessage = "Invalid request line (method)";
                return;
            }

            _uri = parts[1];

            var ver = parts[2];
            if (ver.Length != 8 ||
                !ver.StartsWith("HTTP/") ||
                !tryCreateVersion(ver.Substring(5), out _version) ||
                _version.Major < 1)
            {
                _context.ErrorMessage = "Invalid request line (version)";
            }
        }

        public override string ToString()
        {
            var buff = new StringBuilder(64);
            buff.AppendFormat("{0} {1} HTTP/{2}\r\n", HttpMethod, _uri, _version);
            buff.Append(_headers.ToString());

            return buff.ToString();
        }
    }
}