// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.IO;
using System.Security.Principal;
using System.Text;
using System.Threading;

namespace EonaCat.Network
{
    public class HttpServer
    {
        private string _docRootPath;
        private string _hostname;
        private HttpListener _listener;
        private Thread _receiveThread;
        private volatile ServerState _state;
        private object _sync;

        public HttpServer()
        {
            init("*", System.Net.IPAddress.Any, 80, false);
        }

        public HttpServer(int port)
          : this(port, port == 443)
        {
        }

        public HttpServer(string url)
        {
            if (url == null)
            {
                throw new ArgumentNullException(nameof(url));
            }

            if (url.Length == 0)
            {
                throw new ArgumentException("An empty string.", nameof(url));
            }

            if (!tryCreateUri(url, out Uri uri, out string message))
            {
                throw new ArgumentException(message, nameof(url));
            }

            var host = uri.GetDnsSafeHost(true);

            var addr = host.ToIPAddress();
            if (addr == null)
            {
                message = "The host part could not be converted to an IP address.";
                throw new ArgumentException(message, nameof(url));
            }

            if (!addr.IsLocal())
            {
                message = "The IP address of the host is not a local IP address.";
                throw new ArgumentException(message, nameof(url));
            }

            init(host, addr, uri.Port, uri.Scheme == "https");
        }

        public HttpServer(int port, bool secure)
        {
            if (!port.IsPortNumber())
            {
                var message = "Less than 1 or greater than 65535.";
                throw new ArgumentOutOfRangeException(nameof(port), message);
            }

            init("*", System.Net.IPAddress.Any, port, secure);
        }

        public HttpServer(System.Net.IPAddress address, int port)
          : this(address, port, port == 443)
        {
        }

        public HttpServer(System.Net.IPAddress address, int port, bool secure)
        {
            if (address == null)
            {
                throw new ArgumentNullException(nameof(address));
            }

            if (!address.IsLocal())
            {
                throw new ArgumentException("Not a local IP address.", nameof(address));
            }

            if (!port.IsPortNumber())
            {
                var message = "Less than 1 or greater than 65535.";
                throw new ArgumentOutOfRangeException(nameof(port), message);
            }

            init(address.ToString(true), address, port, secure);
        }

        public System.Net.IPAddress Address { get; private set; }

        public AuthenticationSchemes AuthenticationSchemes
        {
            get
            {
                return _listener.AuthenticationSchemes;
            }

            set
            {
                if (!canSet(out string message))
                {
                    Logger.Warning(message);
                    return;
                }

                lock (_sync)
                {
                    if (!canSet(out message))
                    {
                        Logger.Warning(message);
                        return;
                    }

                    _listener.AuthenticationSchemes = value;
                }
            }
        }

        public string DocumentRootPath
        {
            get
            {
                return _docRootPath;
            }

            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException(nameof(value));
                }

                if (value.Length == 0)
                {
                    throw new ArgumentException("An empty string.", nameof(value));
                }

                value = value.TrimSlashOrBackslashFromEnd();

                string full = null;
                try
                {
                    full = Path.GetFullPath(value);
                }
                catch (Exception ex)
                {
                    throw new ArgumentException("An invalid path string.", nameof(value), ex);
                }

                if (value == "/")
                {
                    throw new ArgumentException("An absolute root.", nameof(value));
                }

                if (value == "\\")
                {
                    throw new ArgumentException("An absolute root.", nameof(value));
                }

                if (value.Length == 2 && value[1] == ':')
                {
                    throw new ArgumentException("An absolute root.", nameof(value));
                }

                if (full == "/")
                {
                    throw new ArgumentException("An absolute root.", nameof(value));
                }

                full = full.TrimSlashOrBackslashFromEnd();
                if (full.Length == 2 && full[1] == ':')
                {
                    throw new ArgumentException("An absolute root.", nameof(value));
                }

                if (!canSet(out string message))
                {
                    Logger.Warning(message);
                    return;
                }

                lock (_sync)
                {
                    if (!canSet(out message))
                    {
                        Logger.Warning(message);
                        return;
                    }

                    _docRootPath = value;
                }
            }
        }

        public bool IsListening => _state == ServerState.Start;

        public bool IsSecure { get; private set; }

        public bool KeepClean
        {
            get
            {
                return WebSocketServices.AutoCleanSessions;
            }

            set
            {
                WebSocketServices.AutoCleanSessions = value;
            }
        }

        public bool IsLoggingEnabled { get; private set; }

        public int Port { get; private set; }

        public string Realm
        {
            get
            {
                return _listener.Realm;
            }

            set
            {
                if (!canSet(out string message))
                {
                    Logger.Warning(message);
                    return;
                }

                lock (_sync)
                {
                    if (!canSet(out message))
                    {
                        Logger.Warning(message);
                        return;
                    }

                    _listener.Realm = value;
                }
            }
        }

        public bool ReuseAddress
        {
            get
            {
                return _listener.ReuseAddress;
            }

            set
            {
                if (!canSet(out string message))
                {
                    Logger.Warning(message);
                    return;
                }

                lock (_sync)
                {
                    if (!canSet(out message))
                    {
                        Logger.Warning(message);
                        return;
                    }

                    _listener.ReuseAddress = value;
                }
            }
        }

        public SSLConfigServer SslConfiguration
        {
            get
            {
                if (!IsSecure)
                {
                    var message = "This instance does not provide secure connections.";
                    throw new InvalidOperationException(message);
                }

                return _listener.SslConfiguration;
            }
        }

        public Func<IIdentity, NetworkCredential> UserCredentialsFinder
        {
            get
            {
                return _listener.UserCredentialsFinder;
            }

            set
            {
                if (!canSet(out string message))
                {
                    Logger.Warning(message);
                    return;
                }

                lock (_sync)
                {
                    if (!canSet(out message))
                    {
                        Logger.Warning(message);
                        return;
                    }

                    _listener.UserCredentialsFinder = value;
                }
            }
        }

        public TimeSpan WaitTime
        {
            get
            {
                return WebSocketServices.WaitTime;
            }

            set
            {
                WebSocketServices.WaitTime = value;
            }
        }

        public WSEndpointManager WebSocketServices { get; private set; }

        public event EventHandler<HttpRequestEventArgs> OnConnect;

        public event EventHandler<HttpRequestEventArgs> OnDelete;

        public event EventHandler<HttpRequestEventArgs> OnGet;

        public event EventHandler<HttpRequestEventArgs> OnHead;

        public event EventHandler<HttpRequestEventArgs> OnOptions;

        public event EventHandler<HttpRequestEventArgs> OnPatch;

        public event EventHandler<HttpRequestEventArgs> OnPost;

        public event EventHandler<HttpRequestEventArgs> OnPut;

        public event EventHandler<HttpRequestEventArgs> OnTrace;

        private void abort()
        {
            lock (_sync)
            {
                if (_state != ServerState.Start)
                {
                    return;
                }

                _state = ServerState.ShuttingDown;
            }

            try
            {
                try
                {
                    WebSocketServices.Stop((ushort)CloseStatusCode.Abnormal, string.Empty);
                }
                finally
                {
                    _listener.Abort();
                }
            }
            catch
            {
            }

            _state = ServerState.Stop;
        }

        private bool canSet(out string message)
        {
            message = null;

            if (_state == ServerState.Start)
            {
                message = "The server has already started.";
                return false;
            }

            if (_state == ServerState.ShuttingDown)
            {
                message = "The server is shutting down.";
                return false;
            }

            return true;
        }

        private bool checkCertificate(out string message)
        {
            message = null;

            var byUser = _listener.SslConfiguration.Certificate != null;

            var path = _listener.CertificateFolderPath;
            var withPort = EndPointListener.CertificateExists(Port, path);

            var both = byUser && withPort;
            if (both)
            {
                Logger.Warning("A server certificate associated with the port is used.");
                return true;
            }

            var either = byUser || withPort;
            if (!either)
            {
                message = "There is no server certificate for secure connections.";
                return false;
            }

            return true;
        }

        private static HttpListener createListener(
          string hostname, int port, bool secure
        )
        {
            var lsnr = new HttpListener();

            var schm = secure ? "https" : "http";
            var pref = string.Format("{0}://{1}:{2}/", schm, hostname, port);
            lsnr.Prefixes.Add(pref);

            return lsnr;
        }

        private void init(
          string hostname, System.Net.IPAddress address, int port, bool secure
        )
        {
            _hostname = hostname;
            Address = address;
            Port = port;
            IsSecure = secure;

            _docRootPath = "./Public";
            _listener = createListener(_hostname, Port, IsSecure);
            WebSocketServices = new WSEndpointManager();
            _sync = new object();
        }

        private void processRequest(HttpListenerContext context)
        {
            var method = context.Request.HttpMethod;
            var evt = method == "GET"
                      ? OnGet
                      : method == "HEAD"
                        ? OnHead
                        : method == "POST"
                          ? OnPost
                          : method == "PUT"
                            ? OnPut
                            : method == "DELETE"
                              ? OnDelete
                              : method == "OPTIONS"
                                ? OnOptions
                                : method == "TRACE"
                                  ? OnTrace
                                  : method == "CONNECT"
                                    ? OnConnect
                                    : method == "PATCH"
                                      ? OnPatch
                                      : null;

            if (evt != null)
            {
                evt(this, new HttpRequestEventArgs(context, _docRootPath));
            }
            else
            {
                context.Response.StatusCode = 501; // Not Implemented
            }

            context.Response.Close();
        }

        private void processRequest(HttpListenerWSContext context)
        {
            var path = context.RequestUri.AbsolutePath;

            if (!WebSocketServices.InternalTryGetServiceHost(path, out WSEndpointHost host))
            {
                context.Close(HttpStatusCode.NotImplemented);
                return;
            }

            host.StartSession(context);
        }

        private void receiveRequest()
        {
            while (true)
            {
                HttpListenerContext ctx = null;
                try
                {
                    ctx = _listener.GetContext();
                    ThreadPool.QueueUserWorkItem(
                      state =>
                      {
                          try
                          {
                              if (ctx.Request.IsUpgradeTo("websocket"))
                              {
                                  processRequest(ctx.AcceptWebSocket(null));
                                  return;
                              }

                              processRequest(ctx);
                          }
                          catch (Exception ex)
                          {
                              Logger.Error(ex.Message);
                              Logger.Debug(ex.ToString());

                              ctx.Connection.Close(true);
                          }
                      }
                    );
                }
                catch (HttpListenerException)
                {
                    Logger.Info("The underlying listener is stopped.");
                    break;
                }
                catch (InvalidOperationException)
                {
                    Logger.Info("The underlying listener is stopped.");
                    break;
                }
                catch (Exception ex)
                {
                    Logger.Error(ex.Message);
                    Logger.Debug(ex.ToString());

                    ctx?.Connection.Close(true);

                    break;
                }
            }

            if (_state != ServerState.ShuttingDown)
            {
                abort();
            }
        }

        private void start()
        {
            if (_state == ServerState.Start)
            {
                Logger.Info("The server has already started.");
                return;
            }

            if (_state == ServerState.ShuttingDown)
            {
                Logger.Warning("The server is shutting down.");
                return;
            }

            lock (_sync)
            {
                if (_state == ServerState.Start)
                {
                    Logger.Info("The server has already started.");
                    return;
                }

                if (_state == ServerState.ShuttingDown)
                {
                    Logger.Warning("The server is shutting down.");
                    return;
                }

                WebSocketServices.Start();

                try
                {
                    startReceiving();
                }
                catch
                {
                    WebSocketServices.Stop((ushort)CloseStatusCode.ServerError, string.Empty);
                    throw;
                }

                _state = ServerState.Start;
            }
        }

        private void startReceiving()
        {
            try
            {
                _listener.Start();
            }
            catch (Exception exception)
            {
                var message = "The underlying listener has failed to start.";
                throw new InvalidOperationException(message, exception);
            }

            _receiveThread = new Thread(new ThreadStart(receiveRequest));
            _receiveThread.IsBackground = true;
            _receiveThread.Start();
        }

        private void stop(ushort code, string reason)
        {
            if (_state == ServerState.Started)
            {
                Logger.Info("The server is not started.");
                return;
            }

            if (_state == ServerState.ShuttingDown)
            {
                Logger.Info("The server is shutting down.");
                return;
            }

            if (_state == ServerState.Stop)
            {
                Logger.Info("The server has already stopped.");
                return;
            }

            lock (_sync)
            {
                if (_state == ServerState.ShuttingDown)
                {
                    Logger.Info("The server is shutting down.");
                    return;
                }

                if (_state == ServerState.Stop)
                {
                    Logger.Info("The server has already stopped.");
                    return;
                }

                _state = ServerState.ShuttingDown;
            }

            try
            {
                var threw = false;
                try
                {
                    WebSocketServices.Stop(code, reason);
                }
                catch
                {
                    threw = true;
                    throw;
                }
                finally
                {
                    try
                    {
                        stopReceiving(5000);
                    }
                    catch
                    {
                        if (!threw)
                        {
                            throw;
                        }
                    }
                }
            }
            finally
            {
                _state = ServerState.Stop;
            }
        }

        private void stopReceiving(int millisecondsTimeout)
        {
            _listener.Stop();
            _receiveThread.Join(millisecondsTimeout);
        }

        private static bool tryCreateUri(
          string uriString, out Uri result, out string message
        )
        {
            result = null;
            message = null;

            var uri = uriString.ToUri();
            if (uri == null)
            {
                message = "An invalid URI string.";
                return false;
            }

            if (!uri.IsAbsoluteUri)
            {
                message = "A relative URI.";
                return false;
            }

            var schm = uri.Scheme;
            if (!(schm == "http" || schm == "https"))
            {
                message = "The scheme part is not 'http' or 'https'.";
                return false;
            }

            if (uri.PathAndQuery != "/")
            {
                message = "It includes either or both path and query components.";
                return false;
            }

            if (uri.Fragment.Length > 0)
            {
                message = "It includes the fragment component.";
                return false;
            }

            if (uri.Port == 0)
            {
                message = "The port part is zero.";
                return false;
            }

            result = uri;
            return true;
        }

        public void AddWebSocketService<TEndpoint>(string path)
          where TEndpoint : WSEndpoint, new()
        {
            WebSocketServices.AddService<TEndpoint>(path, null);
        }

        public void AddWebSocketService<TEndpoint>(
          string path, Action<TEndpoint> initializer
        )
          where TEndpoint : WSEndpoint, new()
        {
            WebSocketServices.AddService(path, initializer);
        }

        public bool RemoveWebSocketService(string path)
        {
            return WebSocketServices.RemoveService(path);
        }

        public void Start()
        {
            if (IsSecure)
            {
                if (!checkCertificate(out string message))
                {
                    throw new InvalidOperationException(message);
                }
            }

            start();
        }

        public void Stop()
        {
            stop((ushort)CloseStatusCode.NoStatus, string.Empty);
        }

        public void Stop(ushort code, string reason)
        {
            if (!code.IsCloseStatusCode())
            {
                var message = "Less than 1000 or greater than 4999.";
                throw new ArgumentOutOfRangeException(nameof(code), message);
            }

            if (code == (ushort)CloseStatusCode.MandatoryExtension)
            {
                var message = $"{(ushort)CloseStatusCode.MandatoryExtension} cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!reason.IsNullOrEmpty())
            {
                if (code == (ushort)CloseStatusCode.NoStatus)
                {
                    var message = $"{(ushort)CloseStatusCode.NoStatus} cannot be used.";
                    throw new ArgumentException(message, nameof(code));
                }

                if (!reason.TryGetUTF8EncodedBytes(out byte[] bytes))
                {
                    var message = "It could not be UTF-8-encoded.";
                    throw new ArgumentException(message, nameof(reason));
                }

                if (bytes.Length > 123)
                {
                    var message = "Its size is greater than 123 bytes.";
                    throw new ArgumentOutOfRangeException(nameof(reason), message);
                }
            }

            stop(code, reason);
        }

        public void Stop(CloseStatusCode code, string reason)
        {
            if (code == CloseStatusCode.MandatoryExtension)
            {
                var message = "MandatoryExtension cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!reason.IsNullOrEmpty())
            {
                if (code == CloseStatusCode.NoStatus)
                {
                    var message = "NoStatus cannot be used.";
                    throw new ArgumentException(message, nameof(code));
                }

                if (!reason.TryGetUTF8EncodedBytes(out byte[] bytes))
                {
                    var message = "It could not be UTF-8-encoded.";
                    throw new ArgumentException(message, nameof(reason));
                }

                if (bytes.Length > 123)
                {
                    var message = "Its size is greater than 123 bytes.";
                    throw new ArgumentOutOfRangeException(nameof(reason), message);
                }
            }

            stop((ushort)code, reason);
        }
    }
}