// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections.Specialized;
using System.IO;
using System.Reflection;
using System.Text;

namespace EonaCat.Network
{
    internal class WebResponse : WebBase
    {
        private WebResponse(string code, string reason, Version version, NameValueCollection headers)
          : base(version, headers)
        {
            StatusCode = code;
            Reason = reason;
        }

        internal WebResponse(HttpStatusCode code)
          : this(code, code.GetDescription())
        {
        }

        internal WebResponse(HttpStatusCode code, string reason)
          : this(((int)code).ToString(), reason, HttpVersion.Version11, new NameValueCollection())
        {
            Headers["Server"] = $"EonaCat.Network/{Constants.Version}";
        }

        public CookieCollection Cookies => Headers.GetCookies(true);

        public bool HasConnectionClose => Headers.Contains("Connection", "close");

        public bool IsProxyAuthenticationRequired => StatusCode == "407";

        public bool IsRedirect => StatusCode == "301" || StatusCode == "302";

        public bool IsUnauthorized => StatusCode == "401";

        public bool IsWebSocketResponse
        {
            get
            {
                var headers = Headers;
                return ProtocolVersion > HttpVersion.Version10 &&
                       StatusCode == "101" &&
                       headers.Contains("Upgrade", "websocket") &&
                       headers.Contains("Connection", "Upgrade");
            }
        }

        public string Reason { get; }

        public string StatusCode { get; }

        internal static WebResponse CreateCloseResponse(HttpStatusCode code)
        {
            var res = new WebResponse(code);
            res.Headers["Connection"] = "close";

            return res;
        }

        internal static WebResponse CreateUnauthorizedResponse(string challenge)
        {
            var res = new WebResponse(HttpStatusCode.Unauthorized);
            res.Headers["WWW-Authenticate"] = challenge;

            return res;
        }

        internal static WebResponse CreateWebSocketResponse()
        {
            var res = new WebResponse(HttpStatusCode.SwitchingProtocols);

            var headers = res.Headers;
            headers["Upgrade"] = "websocket";
            headers["Connection"] = "Upgrade";

            return res;
        }

        internal static WebResponse Parse(string[] headerParts)
        {
            var statusLine = headerParts[0].Split(new[] { ' ' }, 3);
            if (statusLine.Length != 3)
            {
                throw new ArgumentException("Invalid status line: " + headerParts[0]);
            }

            var headers = new WebHeaderCollection();
            for (int i = 1; i < headerParts.Length; i++)
            {
                headers.InternalSet(headerParts[i], true);
            }

            return new WebResponse(
              statusLine[1], statusLine[2], new Version(statusLine[0].Substring(5)), headers);
        }

        internal static WebResponse Read(Stream stream, int millisecondsTimeout)
        {
            return Read(stream, Parse, millisecondsTimeout);
        }

        public void SetCookies(CookieCollection cookies)
        {
            if (cookies == null || cookies.Count == 0)
            {
                return;
            }

            var headers = Headers;
            foreach (var cookie in cookies.Sorted)
            {
                headers.Add("Set-Cookie", cookie.ToResponseString());
            }
        }

        public override string ToString()
        {
            var output = new StringBuilder(64);
            output.AppendFormat("HTTP/{0} {1} {2}{3}", ProtocolVersion, StatusCode, Reason, CrLf);

            var headers = Headers;
            foreach (var key in headers.AllKeys)
            {
                output.AppendFormat("{0}: {1}{2}", key, headers[key], CrLf);
            }

            output.Append(CrLf);

            var entity = EntityBody;
            if (entity.Length > 0)
            {
                output.Append(entity);
            }

            return output.ToString();
        }
    }
}